from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K


class Option(Enum):
    lagrangian = "lagrangian"
    milp = "milp"


class Slope(Enum):
    V_SLOPE = "volume-slope"
    A_SLOPE = "adaptative-slope"
    S_SLOPE = "same-slope"
    Z_SLOPE = "zero-lb"
    O_SLOPE = "one-lb"


class PerturbationDomain(ABC):
    opt_option: Option

    def __init__(self, opt_option: Union[str, Option] = Option.milp):
        self.opt_option = Option(opt_option)

    @abstractmethod
    def get_upper(self, x: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
        ...

    @abstractmethod
    def get_lower(self, x: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
        ...

    @abstractmethod
    def get_nb_x_components(self) -> int:
        ...

    def get_x_input_shape(self, original_input_dim: int) -> Tuple[int, ...]:
        n_comp_x = self.get_nb_x_components()
        if n_comp_x == 1:
            return (original_input_dim,)
        else:
            return (
                n_comp_x,
                original_input_dim,
            )


class BoxDomain(PerturbationDomain):
    def get_upper(self, x: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
        x_min = x[:, 0]
        x_max = x[:, 1]
        return get_upper_box(x_min=x_min, x_max=x_max, w=w, b=b, **kwargs)

    def get_lower(self, x: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
        x_min = x[:, 0]
        x_max = x[:, 1]
        return get_lower_box(x_min=x_min, x_max=x_max, w=w, b=b, **kwargs)

    def get_nb_x_components(self) -> int:
        return 2


class GridDomain(PerturbationDomain):
    pass


class VertexDomain(PerturbationDomain):
    pass


class BallDomain(PerturbationDomain):
    def __init__(self, eps: float, p: float = 2, opt_option: Option = Option.milp):
        super().__init__(opt_option=opt_option)
        self.eps = eps
        # check on p
        p_error_msg = "p must be a positive integer or np.inf"
        try:
            if p != np.inf and (int(p) != p or p <= 0):
                raise ValueError(p_error_msg)
        except:
            raise ValueError(p_error_msg)
        self.p = p

    def get_lower(self, x: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
        return get_lower_ball(x=x, eps=self.eps, p=self.p, w=w, b=b, **kwargs)

    def get_upper(self, x: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
        return get_upper_ball(x=x, eps=self.eps, p=self.p, w=w, b=b, **kwargs)

    def get_nb_x_components(self) -> int:
        return 1


class ForwardMode(Enum):
    """The different forward (from input to output) linear based relaxation perturbation analysis."""

    IBP = "ibp"
    """Propagation of constant bounds from input to output."""

    AFFINE = "affine"
    """Propagation of affine bounds from input to output."""

    HYBRID = "hybrid"
    """Propagation of constant and affines bounds from input to output."""


def get_mode(ibp: bool = True, affine: bool = True) -> ForwardMode:
    if ibp:
        if affine:
            return ForwardMode.HYBRID
        else:
            return ForwardMode.IBP
    else:
        return ForwardMode.AFFINE


def get_ibp(mode: Union[str, ForwardMode] = ForwardMode.HYBRID) -> bool:
    mode = ForwardMode(mode)
    if mode in [ForwardMode.HYBRID, ForwardMode.IBP]:
        return True
    return False


def get_affine(mode: Union[str, ForwardMode] = ForwardMode.HYBRID) -> bool:
    mode = ForwardMode(mode)
    if mode in [ForwardMode.HYBRID, ForwardMode.AFFINE]:
        return True
    return False


class InputsOutputsSpec:
    """Storing specifications for inputs and outputs of decomon/backward layer/model."""

    def __init__(
        self,
        dc_decomp: bool = False,
        mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
        perturbation_domain: Optional[PerturbationDomain] = None,
    ):
        """
        Args:
            dc_decomp: boolean that indicates whether we return a
                difference of convex decomposition of our layer
            mode: type of Forward propagation (ibp, affine, or hybrid)
            perturbation_domain: type of perturbation domain (box, ball, ...)

        """

        self.mode = ForwardMode(mode)
        self.dc_decomp = dc_decomp
        self.perturbation_domain: PerturbationDomain
        if perturbation_domain is None:
            self.perturbation_domain = BoxDomain()
        else:
            self.perturbation_domain = perturbation_domain

    @property
    def nb_tensors(self) -> int:
        if self.mode == ForwardMode.HYBRID:
            nb_tensors = 7
        elif self.mode == ForwardMode.IBP:
            nb_tensors = 2
        elif self.mode == ForwardMode.AFFINE:
            nb_tensors = 5
        else:
            raise NotImplementedError(f"unknown forward mode {self.mode}")

        if self.dc_decomp:
            nb_tensors += 2

        return nb_tensors

    @property
    def ibp(self) -> bool:
        return get_ibp(self.mode)

    @property
    def affine(self) -> bool:
        return get_affine(self.mode)

    def get_input_shape(self, inputsformode: List[tf.Tensor]) -> tf.TensorShape:
        return inputsformode[-1].shape

    def get_fullinputs_from_inputsformode(
        self, inputsformode: List[tf.Tensor], compute_ibp_from_affine: bool = True, tight: bool = True
    ) -> List[tf.Tensor]:
        """

        Args:
            inputsformode:
            compute_ibp_from_affine: if True and mode == affine, compute ibp bounds from affine ones
                with get_upper/get_lower
            tight: if True and mode==hybrid, compute tight ibp bounds, i.e. take tighter bound between
                - the ones from inputs, and
                - the ones computed from affine bounds with get_upper/get_lower

        Returns:

        """
        dtype = inputsformode[0].dtype
        nb_tensors = self.nb_tensors
        nonelike_tensor = self.get_empty_tensor(dtype=dtype)
        if self.dc_decomp:
            if self.mode == ForwardMode.HYBRID:
                x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputsformode[:nb_tensors]
            elif self.mode == ForwardMode.IBP:
                u_c, l_c, h, g = inputsformode[:nb_tensors]
                x, w_u, b_u, w_l, b_l = (
                    nonelike_tensor,
                    nonelike_tensor,
                    nonelike_tensor,
                    nonelike_tensor,
                    nonelike_tensor,
                )
            elif self.mode == ForwardMode.AFFINE:
                x, w_u, b_u, w_l, b_l, h, g = inputsformode[:nb_tensors]
                u_c, l_c = nonelike_tensor, nonelike_tensor
            else:
                raise ValueError(f"Unknown mode {self.mode}")
        else:
            h, g = nonelike_tensor, nonelike_tensor
            if self.mode == ForwardMode.HYBRID:
                x, u_c, w_u, b_u, l_c, w_l, b_l = inputsformode[:nb_tensors]
            elif self.mode == ForwardMode.IBP:
                u_c, l_c = inputsformode[:nb_tensors]
                x, w_u, b_u, w_l, b_l = (
                    nonelike_tensor,
                    nonelike_tensor,
                    nonelike_tensor,
                    nonelike_tensor,
                    nonelike_tensor,
                )
            elif self.mode == ForwardMode.AFFINE:
                x, w_u, b_u, w_l, b_l = inputsformode[:nb_tensors]
                u_c, l_c = nonelike_tensor, nonelike_tensor
            else:
                raise ValueError(f"Unknown mode {self.mode}")

        compute_ibp_from_affine = (compute_ibp_from_affine and self.mode == ForwardMode.AFFINE) or (
            tight and self.mode == ForwardMode.HYBRID
        )

        if compute_ibp_from_affine:
            u_c_affine = self.perturbation_domain.get_upper(x, w_u, b_u)
            l_c_affine = self.perturbation_domain.get_lower(x, w_l, b_l)
            if self.mode == ForwardMode.AFFINE:
                u_c = u_c_affine
                l_c = l_c_affine
            else:
                u_c = K.minimum(u_c, u_c_affine)
                l_c = K.maximum(l_c, l_c_affine)

        return [x, u_c, w_u, b_u, l_c, w_l, b_l, h, g]

    def get_fullinputs_by_type_from_inputsformode_to_merge(
        self, inputsformode: List[tf.Tensor], compute_ibp_from_affine: bool = False, tight: bool = True
    ) -> List[List[tf.Tensor]]:
        """

        Args:
            inputsformode:
            compute_ibp_from_affine: if True and mode == affine, compute ibp bounds from affine ones
                with get_upper/get_lower
            tight: if True and mode==hybrid, compute tight ibp bounds, i.e. take tighter bound between
                - the ones from inputs, and
                - the ones computed from affine bounds with get_upper/get_lower

        Returns:

        """
        dtype = inputsformode[0].dtype
        nb_tensors_by_input = self.nb_tensors
        nb_inputs = len(inputsformode) // nb_tensors_by_input
        nonelike_tensor = self.get_empty_tensor(dtype=dtype)
        nonelike_tensor_list = [nonelike_tensor] * nb_inputs
        if self.mode == ForwardMode.HYBRID:
            inputs_x = inputsformode[0::nb_tensors_by_input]
            inputs_u_c = inputsformode[1::nb_tensors_by_input]
            inputs_w_u = inputsformode[2::nb_tensors_by_input]
            inputs_b_u = inputsformode[3::nb_tensors_by_input]
            inputs_l_c = inputsformode[4::nb_tensors_by_input]
            inputs_w_l = inputsformode[5::nb_tensors_by_input]
            inputs_b_l = inputsformode[6::nb_tensors_by_input]
        elif self.mode == ForwardMode.IBP:
            inputs_u_c = inputsformode[0::nb_tensors_by_input]
            inputs_l_c = inputsformode[1::nb_tensors_by_input]
            inputs_x, inputs_w_u, inputs_b_u, inputs_w_l, inputs_b_l = (
                nonelike_tensor_list,
                nonelike_tensor_list,
                nonelike_tensor_list,
                nonelike_tensor_list,
                nonelike_tensor_list,
            )
        elif self.mode == ForwardMode.AFFINE:
            inputs_x = inputsformode[0::nb_tensors_by_input]
            inputs_w_u = inputsformode[1::nb_tensors_by_input]
            inputs_b_u = inputsformode[2::nb_tensors_by_input]
            inputs_w_l = inputsformode[3::nb_tensors_by_input]
            inputs_b_l = inputsformode[4::nb_tensors_by_input]
            inputs_u_c, inputs_l_c = nonelike_tensor_list, nonelike_tensor_list
        else:
            raise ValueError(f"Unknown mode {self.mode}")

        if self.dc_decomp:
            inputs_h = inputsformode[nb_tensors_by_input - 2 :: nb_tensors_by_input]
            inputs_g = inputsformode[nb_tensors_by_input - 1 :: nb_tensors_by_input]
        else:
            inputs_h, inputs_g = nonelike_tensor_list, nonelike_tensor_list

        # compute ibp bounds from affine bounds
        compute_ibp_from_affine = (compute_ibp_from_affine and self.mode == ForwardMode.AFFINE) or (
            tight and self.mode == ForwardMode.HYBRID
        )

        if compute_ibp_from_affine:
            for i in range(len(inputs_x)):
                u_c_affine = self.perturbation_domain.get_upper(inputs_x[i], inputs_w_u[i], inputs_b_u[i])
                l_c_affine = self.perturbation_domain.get_lower(inputs_x[i], inputs_w_l[i], inputs_b_l[i])
                if self.mode == ForwardMode.AFFINE:
                    inputs_u_c[i] = u_c_affine
                    inputs_l_c[i] = l_c_affine
                else:
                    inputs_u_c[i] = K.minimum(inputs_u_c[i], u_c_affine)
                    inputs_l_c[i] = K.maximum(inputs_l_c[i], l_c_affine)

        return [inputs_x, inputs_u_c, inputs_w_u, inputs_b_u, inputs_l_c, inputs_w_l, inputs_b_l, inputs_h, inputs_g]

    def split_inputsformode_to_merge(self, inputsformode: List[tf.Tensor]) -> List[List[tf.Tensor]]:
        n_comp = self.nb_tensors
        return [inputsformode[n_comp * i : n_comp * (i + 1)] for i in range(len(inputsformode) // n_comp)]

    def extract_inputsformode_from_fullinputs(self, inputs: List[tf.Tensor]) -> List[tf.Tensor]:
        x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs
        if self.mode == ForwardMode.HYBRID:
            inputsformode = [x, u_c, w_u, b_u, l_c, w_l, b_l]
        elif self.mode == ForwardMode.IBP:
            inputsformode = [u_c, l_c]
        elif self.mode == ForwardMode.AFFINE:
            inputsformode = [x, w_u, b_u, w_l, b_l]
        else:
            raise ValueError(f"Unknown mode {self.mode}")
        if self.dc_decomp:
            inputsformode += [h, g]
        return inputsformode

    def extract_outputsformode_from_fulloutputs(self, outputs: List[tf.Tensor]) -> List[tf.Tensor]:
        return self.extract_inputsformode_from_fullinputs(outputs)

    @staticmethod
    def get_empty_tensor(dtype: Union[str, tf.DType] = K.floatx()) -> tf.Tensor:
        return tf.constant([], dtype=dtype)


def get_upper_box(x_min: tf.Tensor, x_max: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
    """#compute the max of an affine function
    within a box (hypercube) defined by its extremal corners

    Args:
        x_min: lower bound of the box domain
        x_max: upper bound of the box domain
        w: weights of the affine function
        b: bias of the affine function

    Returns:
        max_(x >= x_min, x<=x_max) w*x + b
    """

    if len(w.shape) == len(b.shape):  # identity function
        return x_max

    # split into positive and negative components
    z_value = K.cast(0.0, dtype=x_min.dtype)
    w_pos = K.maximum(w, z_value)
    w_neg = K.minimum(w, z_value)

    x_min_out = x_min + z_value * x_min
    x_max_out = x_max + z_value * x_max

    for _ in range(len(w.shape) - len(x_max.shape)):
        x_min_out = K.expand_dims(x_min_out, -1)
        x_max_out = K.expand_dims(x_max_out, -1)

    return K.sum(w_pos * x_max_out + w_neg * x_min_out, 1) + b


def get_lower_box(x_min: tf.Tensor, x_max: tf.Tensor, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
    """
    Args:
        x_min: lower bound of the box domain
        x_max: upper bound of the box domain
        w_l: weights of the affine lower bound
        b_l: bias of the affine lower bound

    Returns:
        min_(x >= x_min, x<=x_max) w*x + b
    """

    if len(w.shape) == len(b.shape):
        return x_min

    z_value = K.cast(0.0, dtype=x_min.dtype)

    w_pos = K.maximum(w, z_value)
    w_neg = K.minimum(w, z_value)

    x_min_out = x_min + z_value * x_min
    x_max_out = x_max + z_value * x_max

    for _ in range(len(w.shape) - len(x_max.shape)):
        x_min_out = K.expand_dims(x_min_out, -1)
        x_max_out = K.expand_dims(x_max_out, -1)

    return K.sum(w_pos * x_min_out + w_neg * x_max_out, 1) + b


def get_lq_norm(x: tf.Tensor, p: float, axis: int = -1) -> tf.Tensor:
    """compute Lp norm (p=1 or 2)

    Args:
        x: tensor
        p: the power must be an integer in (1, 2)
        axis: the axis on which we compute the norm

    Returns:
        ||w||^p
    """
    if p == 1:
        x_q = K.max(K.abs(x), axis)
    elif p == 2:
        x_q = K.sqrt(K.sum(K.pow(x, p), axis))
    else:
        raise NotImplementedError("p must be equal to 1 or 2")

    return x_q


def get_upper_ball(x_0: tf.Tensor, eps: float, p: float, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
    """max of an affine function over an Lp ball

    Args:
        x_0: the center of the ball
        eps: the radius
        p: the type of Lp norm considered
        w: weights of the affine function
        b: bias of the affine function

    Returns:
        max_(|x - x_0|_p<= eps) w*x + b
    """
    if len(w.shape) == len(b.shape):
        return x_0 + eps

    if p == np.inf:
        # compute x_min and x_max according to eps
        x_min = x_0 - eps
        x_max = x_0 + eps
        return get_upper_box(x_min, x_max, w, b)

    else:
        # use Holder's inequality p+q=1
        # ||w||_q*eps + w*x_0 + b

        if len(kwargs):
            return get_upper_ball_finetune(x_0, eps, p, w, b, **kwargs)

        upper = eps * get_lq_norm(w, p, axis=1) + b

        for _ in range(len(w.shape) - len(x_0.shape)):
            x_0 = K.expand_dims(x_0, -1)

        return K.sum(w * x_0, 1) + upper


def get_lower_ball(x_0: tf.Tensor, eps: float, p: float, w: tf.Tensor, b: tf.Tensor, **kwargs: Any) -> tf.Tensor:
    """min of an affine fucntion over an Lp ball

    Args:
        x_0: the center of the ball
        eps: the radius
        p: the type of Lp norm considered
        w: weights of the affine function
        b: bias of the affine function

    Returns:
        min_(|x - x_0|_p<= eps) w*x + b
    """
    if len(w.shape) == len(b.shape):
        return x_0 - eps

    if p == np.inf:
        # compute x_min and x_max according to eps
        x_min = x_0 - eps
        x_max = x_0 + eps
        return get_lower_box(x_min, x_max, w, b)

    else:
        # use Holder's inequality p+q=1
        # ||w||_q*eps + w*x_0 + b

        if len(kwargs):
            return get_lower_ball_finetune(x_0, eps, p, w, b, **kwargs)

        lower = -eps * get_lq_norm(w, p, axis=1) + b

        for _ in range(len(w.shape) - len(x_0.shape)):
            x_0 = K.expand_dims(x_0, -1)

        return K.sum(w * x_0, 1) + lower


def get_lower_ball_finetune(
    x_0: tf.Tensor, eps: float, p: float, w: tf.Tensor, b: tf.Tensor, **kwargs: Any
) -> tf.Tensor:
    if "finetune_lower" in kwargs and "upper" in kwargs or "lower" in kwargs:
        alpha = kwargs["finetune_lower"]
        # assume alpha is the same shape as w, minus the batch dimension
        n_shape = len(w.shape) - 2

        if "upper" and "lower" in kwargs:
            upper = kwargs["upper"]  # flatten vector
            lower = kwargs["lower"]  # flatten vector

            upper_reshaped = np.reshape(upper, [1, -1] + [1] * n_shape)
            lower_reshaped = np.reshape(lower, [1, -1] + [1] * n_shape)

            w_alpha = w * alpha[None]
            w_alpha_bar = w * (1 - alpha)

            score_box = K.sum(K.maximum(0.0, w_alpha_bar) * lower_reshaped, 1) + K.sum(
                K.minimum(0.0, w_alpha_bar) * upper_reshaped, 1
            )
            score_ball = get_lower_ball(x_0, eps, p, w_alpha, b)

            return score_box + score_ball

        if "upper" in kwargs:
            upper = kwargs["upper"]  # flatten vector
            upper_reshaped = np.reshape(upper, [1, -1] + [1] * n_shape)

            w_alpha = K.minimum(0, w) * alpha[None] + K.maximum(0.0, w)
            w_alpha_bar = K.minimum(0, w) * (1 - alpha[None])

            score_box = K.sum(K.minimum(0.0, w_alpha_bar) * upper_reshaped, 1)
            score_ball = get_lower_ball(x_0, eps, p, w_alpha, b)

            return score_box + score_ball

        if "lower" in kwargs:
            lower = kwargs["lower"]  # flatten vector
            lower_reshaped = np.reshape(lower, [1, -1] + [1] * n_shape)

            w_alpha = K.maximum(0, w) * alpha[None] + K.minimum(0.0, w)
            w_alpha_bar = K.maximum(0, w) * (1 - alpha[None])

            score_box = K.sum(K.maximum(0.0, w_alpha_bar) * lower_reshaped, 1)
            score_ball = get_lower_ball(x_0, eps, p, w_alpha, b)

            return score_box + score_ball

    return get_lower_ball(x_0, eps, p, w, b)


def get_upper_ball_finetune(
    x_0: tf.Tensor, eps: float, p: float, w: tf.Tensor, b: tf.Tensor, **kwargs: Any
) -> tf.Tensor:
    if "finetune_upper" in kwargs and "upper" in kwargs or "lower" in kwargs:
        alpha = kwargs["finetune_upper"]
        # assume alpha is the same shape as w, minus the batch dimension
        n_shape = len(w.shape) - 2

        if "upper" and "lower" in kwargs:
            upper = kwargs["upper"]  # flatten vector
            lower = kwargs["lower"]  # flatten vector

            upper_reshaped = np.reshape(upper, [1, -1] + [1] * n_shape)
            lower_reshaped = np.reshape(lower, [1, -1] + [1] * n_shape)

            w_alpha = w * alpha[None]
            w_alpha_bar = w * (1 - alpha)

            score_box = K.sum(K.maximum(0.0, w_alpha_bar) * upper_reshaped, 1) + K.sum(
                K.minimum(0.0, w_alpha_bar) * lower_reshaped, 1
            )
            score_ball = get_lower_ball(x_0, eps, p, w_alpha, b)

            return score_box + score_ball

        if "upper" in kwargs:
            upper = kwargs["upper"]  # flatten vector
            upper_reshaped = np.reshape(upper, [1, -1] + [1] * n_shape)

            w_alpha = K.minimum(0, w) * alpha[None] + K.maximum(0.0, w)
            w_alpha_bar = K.minimum(0, w) * (1 - alpha[None])

            score_box = K.sum(K.maximum(0.0, w_alpha_bar) * upper_reshaped, 1)
            score_ball = get_lower_ball(x_0, eps, p, w_alpha, b)

            return score_box + score_ball

        if "lower" in kwargs:
            lower = kwargs["lower"]  # flatten vector
            lower_reshaped = np.reshape(lower, [1, -1] + [1] * n_shape)

            w_alpha = K.maximum(0, w) * alpha[None] + K.minimum(0.0, w)
            w_alpha_bar = K.maximum(0, w) * (1 - alpha[None])

            score_box = K.sum(K.minimum(0.0, w_alpha_bar) * lower_reshaped, 1)
            score_ball = get_lower_ball(x_0, eps, p, w_alpha, b)

            return score_box + score_ball

    return get_upper_ball(x_0, eps, p, w, b)



from .loss import (
    build_asymptotic_crossentropy_model,
    build_crossentropy_model,
    build_radius_robust_model,
)
from .metric import (
    build_formal_adv_check_model,
    build_formal_adv_model,
    build_formal_upper_model,
)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from tensorflow.keras.layers import Activation, Dense
from tensorflow.keras.models import Sequential

from decomon import get_lower_box, get_range_box, get_upper_box
from decomon.core import get_affine, get_ibp
from decomon.models import clone


@pytest.fixture()
def toy_model_1d():
    sequential = Sequential()
    sequential.add(Dense(1, activation="linear", input_dim=1))
    sequential.add(Activation("relu"))
    sequential.add(Dense(1, activation="linear"))
    return sequential


@pytest.fixture()
def toy_model_multid(odd, helpers):
    input_dim = helpers.get_input_dim_multid_box(odd)
    sequential = Sequential()
    sequential.add(Dense(1, activation="linear", input_dim=input_dim))
    sequential.add(Activation("relu"))
    sequential.add(Dense(1, activation="linear"))
    return sequential


def test_get_upper_1d_box(toy_model_1d, n, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=False)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_

    ibp = get_ibp(mode)
    affine = get_affine(mode)
    backward_model = clone(toy_model_1d, method=method, ibp=ibp, affine=affine, mode=mode)
    upper = get_upper_box(backward_model, z[:, 0], z[:, 1])
    y_ref = toy_model_1d.predict(y)

    try:
        assert (upper - y_ref).min() + 1e-6 >= 0.0
    except AssertionError:
        toy_model_1d.save_weights("get_upper_1d_box_fail_{}_{}_{}.hd5".format(n, method, mode))
        raise AssertionError


def test_get_lower_1d_box(toy_model_1d, n, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=False)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_

    ibp = get_ibp(mode)
    affine = get_affine(mode)
    backward_model = clone(toy_model_1d, method=method, final_ibp=ibp, final_affine=affine)
    lower = get_lower_box(backward_model, z[:, 0], z[:, 1])
    y_ref = toy_model_1d.predict(y)

    try:
        assert (y_ref - lower).min() + 1e-6 >= 0.0
    except AssertionError:
        toy_model_1d.save_weights("get_lower_1d_box_fail_{}_{}_{}.hd5".format(n, method, mode))
        raise AssertionError


def test_get_range_1d_box(toy_model_1d, n, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=False)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_

    ibp = get_ibp(mode)
    affine = get_affine(mode)
    backward_model = clone(toy_model_1d, method=method, final_ibp=ibp, final_affine=affine)
    upper, lower = get_range_box(backward_model, z[:, 0], z[:, 1])
    y_ref = toy_model_1d.predict(y)

    try:
        assert (upper - y_ref).min() + 1e-6 >= 0.0
        assert (y_ref - lower).min() + 1e-6 >= 0.0
    except AssertionError:
        toy_model_1d.save_weights("get_range_1d_box_fail_{}_{}_{}.hd5".format(n, method, mode))
        raise AssertionError


def test_get_upper_multid_box(toy_model_multid, odd, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    inputs_ = helpers.get_standard_values_multid_box_convert(odd, dc_decomp=False)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_
    ibp = get_ibp(mode)
    affine = get_affine(mode)
    backward_model = clone(toy_model_multid, method=method, final_ibp=ibp, final_affine=affine)
    upper = get_upper_box(backward_model, z[:, 0], z[:, 1])
    y_ref = toy_model_multid.predict(y)

    assert (upper - y_ref).min() + 1e-6 >= 0.0


def test_get_lower_multid_box(toy_model_multid, odd, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    inputs_ = helpers.get_standard_values_multid_box_convert(odd, dc_decomp=False)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_

    ibp = get_ibp(mode)
    affine = get_affine(mode)
    backward_model = clone(toy_model_multid, method=method, final_ibp=ibp, final_affine=affine)
    lower = get_lower_box(backward_model, z[:, 0], z[:, 1])
    y_ref = toy_model_multid.predict(y)

    assert (y_ref - lower).min() + 1e-6 >= 0.0


def test_get_range_multid_box(toy_model_multid, odd, method, mode, helpers):
    if not helpers.is_method_mode_compatible(method=method, mode=mode):
        # skip method=ibp/crown-ibp with mode=affine/hybrid
        pytest.skip(f"output mode {mode} is not compatible with convert method {method}")

    inputs_ = helpers.get_standard_values_multid_box_convert(odd, dc_decomp=False)
    x, y, z, u_c, W_u, b_u, l_c, W_l, b_l = inputs_

    ibp = get_ibp(mode)
    affine = get_affine(mode)
    backward_model = clone(toy_model_multid, method=method, final_ibp=ibp, final_affine=affine)
    upper, lower = get_range_box(backward_model, z[:, 0], z[:, 1])
    y_ref = toy_model_multid.predict(y)

    assert (upper - y_ref).min() + 1e-6 >= 0.0
    assert (y_ref - lower).min() + 1e-6 >= 0.0



import numpy as np
import pytest
import tensorflow as tf
import tensorflow.keras.backend as K
from numpy.testing import assert_almost_equal

from decomon.layers.utils_pooling import (
    get_lower_linear_hull_max,
    get_upper_linear_hull_max,
)


@pytest.mark.parametrize(
    "func, minmax, clipmin, clipmax",
    [
        (get_lower_linear_hull_max, np.max, -np.inf, 0.0),
        (get_upper_linear_hull_max, np.min, 0.0, np.inf),
    ],
)
def test_get_lower_upper_linear_hull_max(
    func, minmax, clipmin, clipmax, mode, floatx, decimal, axis, finetune_odd, helpers
):
    if finetune_odd is not None and func is not get_lower_linear_hull_max:
        # skip test with finetune if not get_lower
        pytest.skip("finetune_odd is only intended for get_lower_linear_hull_max()")

    odd, m_0, m_1 = 0, 0, 1
    data_format = "channels_last"
    dc_decomp = True

    # inputs
    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs_)

    if finetune_odd is not None:
        finetune_odd = tf.constant(finetune_odd, dtype="float{}".format(floatx))

    # decomon output
    output = func(inputs_for_mode, mode=mode, axis=axis, finetune_lower=finetune_odd)
    f_pooling = K.function(inputs, output)
    w_, b_ = f_pooling(inputs_)

    # reference output
    output_ref_ = np.max(input_ref_, axis=axis)

    assert_almost_equal(
        minmax(np.clip(np.sum(w_ * input_ref_, axis) + b_ - output_ref_, clipmin, clipmax)),
        0.0,
        decimal=decimal,
        err_msg=f"linear hull for bounding max",
    )



import numpy as np
import pytest
import tensorflow.keras.backend as K
from numpy.testing import assert_almost_equal
from tensorflow.keras.layers import (
    Activation,
    Conv2D,
    Dense,
    Flatten,
    Input,
    InputLayer,
    PReLU,
)
from tensorflow.keras.models import Model, Sequential

from decomon.models.convert import (
    convert_deellip_layers_in_keras_model,
    preprocess_keras_model,
    split_activations_in_keras_model,
)

try:
    import deel.lip
except ImportError:
    deel_lip_available = False
else:
    deel_lip_available = True
    from deel.lip.activations import GroupSort
    from deel.lip.layers import SpectralDense

deel_lip_skip_reason = "deel-lip is not available"


def test_split_activations_in_keras_model_no_inputshape_ko():
    layers = [
        Conv2D(
            10,
            kernel_size=(3, 3),
            activation="relu",
            data_format="channels_last",
        ),
        Flatten(),
        Dense(1),
    ]
    model = Sequential(layers)
    with pytest.raises(ValueError):
        converted_model = split_activations_in_keras_model(model)


def test_split_activations_in_keras_model(toy_model):
    converted_model = split_activations_in_keras_model(toy_model)
    assert isinstance(converted_model, Model)
    # check no more activation functions in non-activation layers
    for layer in converted_model.layers:
        activation = layer.get_config().get("activation", None)
        assert isinstance(layer, Activation) or activation is None or activation == "linear"
    # check same outputs
    input_shape_wo_batchsize = toy_model.input_shape[1:]
    input_shape_with_batch_size = (5,) + input_shape_wo_batchsize
    flatten_dim = np.prod(input_shape_with_batch_size)
    inputs_np = np.linspace(-1, 1, flatten_dim).reshape(input_shape_with_batch_size)
    output_np_ref = toy_model(inputs_np).numpy()
    output_np_new = converted_model(inputs_np).numpy()
    assert_almost_equal(output_np_new, output_np_ref, decimal=4)


def test_convert_deellip_layers_in_keras_model_no_inputshape_ko():
    layers = [
        Conv2D(
            10,
            kernel_size=(3, 3),
            activation="relu",
            data_format="channels_last",
        ),
        Flatten(),
        Dense(1),
    ]
    model = Sequential(layers)
    with pytest.raises(ValueError):
        convert_deellip_layers_in_keras_model(model)


@pytest.mark.skipif(not (deel_lip_available), reason=deel_lip_skip_reason)
def test_convert_deellip_layers_in_keras_model_ok():
    layers = [
        Input(shape=(1,)),
        SpectralDense(3),
        GroupSort(n=1),
        Dense(1),
    ]
    model = Sequential(layers)
    converted_model = convert_deellip_layers_in_keras_model(model)
    # check no more vanilla_export
    for layer in converted_model.layers:
        assert not hasattr(layer, "vanilla_export")

    # check same outputs
    input_shape_wo_batchsize = model.input_shape[1:]
    input_shape_with_batch_size = (5,) + input_shape_wo_batchsize
    flatten_dim = np.prod(input_shape_with_batch_size)
    inputs_np = np.linspace(-1, 1, flatten_dim).reshape(input_shape_with_batch_size)
    output_np_ref = model(inputs_np).numpy()
    output_np_new = converted_model(inputs_np).numpy()
    assert_almost_equal(output_np_new, output_np_ref, decimal=4)


@pytest.mark.parametrize(
    "layer_class_name, "
    "layer_kwargs, "
    "input_shape_wo_batchsize, "
    "embedded_activation_layer_class_name, "
    "embedded_activation_layer_class_kwargs, "
    "is_deellip_layer, "
    "is_deellip_activation",
    [
        ("Dense", {"units": 3, "activation": "relu"}, (1,), None, None, False, False),
        ("Dense", {"units": 3}, (1,), "PReLU", {}, False, False),
        ("Conv2D", {"filters": 2, "kernel_size": (3, 3), "activation": "relu"}, (64, 64, 3), None, None, False, False),
        ("Dense", {"units": 3}, (1,), "GroupSort", {"n": 1}, False, True),
        ("SpectralDense", {"units": 3, "activation": "relu"}, (1,), None, None, True, False),
        ("SpectralDense", {"units": 3}, (1,), "PReLU", {}, True, False),
        ("SpectralDense", {"units": 3}, (1,), "GroupSort", {"n": 1}, True, True),
    ],
)
def test_preprocess(
    layer_class_name,
    layer_kwargs,
    input_shape_wo_batchsize,
    embedded_activation_layer_class_name,
    embedded_activation_layer_class_kwargs,
    use_bias,
    is_deellip_layer,
    is_deellip_activation,
):
    # skip deel-lip cases if not available
    if not deel_lip_available and (is_deellip_layer or is_deellip_activation):
        pytest.skip(reason=deel_lip_skip_reason)
    # init hidden layer
    layer_class = globals()[layer_class_name]
    if embedded_activation_layer_class_name is not None:
        embedded_activation_layer_class = globals()[embedded_activation_layer_class_name]
        embedded_activation_layer = embedded_activation_layer_class(**embedded_activation_layer_class_kwargs)
        hidden_layer = layer_class(use_bias=use_bias, activation=embedded_activation_layer, **layer_kwargs)
    else:
        hidden_layer = layer_class(use_bias=use_bias, **layer_kwargs)
    layers = [
        Input(shape=input_shape_wo_batchsize),
        hidden_layer,
        Dense(1),
    ]
    model = Sequential(layers)
    converted_model = preprocess_keras_model(model)
    # check no more vanilla_export, no more embedded activation
    for layer in converted_model.layers:
        assert not hasattr(layer, "vanilla_export")
        activation = layer.get_config().get("activation", None)
        assert (
            isinstance(layer, Activation)
            or layer.__class__.__name__ == "GroupSort"
            or activation is None
            or activation == "linear"
        )
        if not isinstance(layer, InputLayer):
            assert hasattr(layer, "is_lipschitz")
            assert hasattr(layer, "deellip_classname")
            assert hasattr(layer, "k_coef_lip")
    # check number of layers
    assert len(converted_model.layers) == 4
    # check same outputs
    input_shape_wo_batchsize = model.input_shape[1:]
    input_shape_with_batch_size = (5,) + input_shape_wo_batchsize
    flatten_dim = np.prod(input_shape_with_batch_size)
    inputs_np = np.linspace(-1, 1, flatten_dim).reshape(input_shape_with_batch_size)
    output_np_ref = model(inputs_np).numpy()
    output_np_new = converted_model(inputs_np).numpy()
    assert_almost_equal(output_np_new, output_np_ref, decimal=4)



# extra layers necessary for backward LiRPA
from typing import Any, Dict, List, Optional, Tuple, Union

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.layers import InputSpec, Layer

from decomon.core import ForwardMode, PerturbationDomain


class Fuse(Layer):
    def __init__(self, mode: Union[str, ForwardMode], **kwargs: Any):
        super().__init__(**kwargs)
        self.mode = ForwardMode(mode)

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        inputs_wo_backward_bounds = inputs[:-4]
        backward_bounds = inputs[-4:]

        if self.mode == ForwardMode.AFFINE:
            x_0, w_f_u, b_f_u, w_f_l, b_f_l = inputs_wo_backward_bounds
        elif self.mode == ForwardMode.HYBRID:
            x_0, u_c, w_f_u, b_f_u, l_c, w_f_l, b_f_l = inputs_wo_backward_bounds
        else:
            return backward_bounds

        return merge_with_previous([w_f_u, b_f_u, w_f_l, b_f_l] + backward_bounds)

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update({"mode": self.mode})
        return config


class Convert2BackwardMode(Layer):
    def __init__(self, mode: Union[str, ForwardMode], perturbation_domain: PerturbationDomain, **kwargs: Any):
        super().__init__(**kwargs)
        self.mode = ForwardMode(mode)
        self.perturbation_domain = perturbation_domain

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        inputs_wo_backward_bounds = inputs[:-4]
        backward_bounds = inputs[-4:]
        w_u_out, b_u_out, w_l_out, b_l_out = backward_bounds
        empty_tensor = tf.constant([])

        if self.mode in [ForwardMode.AFFINE, ForwardMode.HYBRID]:
            x_0 = inputs_wo_backward_bounds[0]
        else:
            u_c, l_c = inputs_wo_backward_bounds
            x_0 = K.concatenate([K.expand_dims(l_c, 1), K.expand_dims(u_c, 1)], 1)

        if self.mode in [ForwardMode.IBP, ForwardMode.HYBRID]:
            u_c_out = self.perturbation_domain.get_upper(x_0, w_u_out, b_u_out)
            l_c_out = self.perturbation_domain.get_lower(x_0, w_l_out, b_l_out)
        else:
            u_c_out, l_c_out = empty_tensor, empty_tensor

        if self.mode == ForwardMode.AFFINE:
            return [x_0] + backward_bounds
        elif self.mode == ForwardMode.IBP:
            return [u_c_out, l_c_out]
        elif self.mode == ForwardMode.HYBRID:
            return [x_0, u_c_out, w_u_out, b_u_out, l_c_out, w_l_out, b_l_out]

        else:
            raise ValueError(f"Unknwon mode {self.mode}")

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update({"mode": self.mode, "perturbation_domain": self.perturbation_domain})
        return config


class MergeWithPrevious(Layer):
    def __init__(
        self,
        input_shape_layer: Optional[Tuple[int, ...]] = None,
        backward_shape_layer: Optional[Tuple[int, ...]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.input_shape_layer = input_shape_layer
        self.backward_shape_layer = backward_shape_layer
        if not (input_shape_layer is None) and not (backward_shape_layer is None):
            _, n_in, n_h = input_shape_layer
            _, n_h, n_out = backward_shape_layer

            w_out_spec = InputSpec(ndim=3, axes={-1: n_h, -2: n_in})
            b_out_spec = InputSpec(ndim=2, axes={-1: n_h})
            w_b_spec = InputSpec(ndim=3, axes={-1: n_out, -2: n_h})
            b_b_spec = InputSpec(ndim=2, axes={-1: n_out})
            self.input_spec = [w_out_spec, b_out_spec] * 2 + [w_b_spec, b_b_spec] * 2  #

    def call(self, inputs: List[tf.Tensor], **kwargs: Any) -> List[tf.Tensor]:
        return merge_with_previous(inputs)

    def get_config(self) -> Dict[str, Any]:
        config = super().get_config()
        config.update(
            {
                "input_shape_layer": self.input_shape_layer,
                "backward_shape_layer": self.backward_shape_layer,
            }
        )
        return config


def merge_with_previous(inputs: List[tf.Tensor]) -> List[tf.Tensor]:
    w_u_out, b_u_out, w_l_out, b_l_out, w_b_u, b_b_u, w_b_l, b_b_l = inputs

    # w_u_out (None, n_h_in, n_h_out)
    # w_b_u (None, n_h_out, n_out)

    # w_u_out_ (None, n_h_in, n_h_out, 1)
    # w_b_u_ (None, 1, n_h_out, n_out)
    # w_u_out_*w_b_u_ (None, n_h_in, n_h_out, n_out)

    # result (None, n_h_in, n_out)

    if len(w_u_out.shape) == 2:
        w_u_out = tf.linalg.diag(w_u_out)

    if len(w_l_out.shape) == 2:
        w_l_out = tf.linalg.diag(w_l_out)

    if len(w_b_u.shape) == 2:
        w_b_u = tf.linalg.diag(w_b_u)

    if len(w_b_l.shape) == 2:
        w_b_l = tf.linalg.diag(w_b_l)

    # import pdb; pdb.set_trace()

    z_value = K.cast(0.0, dtype=w_u_out.dtype)
    w_b_u_pos = K.maximum(w_b_u, z_value)
    w_b_u_neg = K.minimum(w_b_u, z_value)
    w_b_l_pos = K.maximum(w_b_l, z_value)
    w_b_l_neg = K.minimum(w_b_l, z_value)

    w_u = K.batch_dot(w_u_out, w_b_u_pos, (-1, -2)) + K.batch_dot(w_l_out, w_b_u_neg, (-1, -2))
    w_l = K.batch_dot(w_l_out, w_b_l_pos, (-1, -2)) + K.batch_dot(w_u_out, w_b_l_neg, (-1, -2))
    b_u = K.batch_dot(b_u_out, w_b_u_pos, (-1, -2)) + K.batch_dot(b_l_out, w_b_u_neg, (-1, -2)) + b_b_u
    b_l = K.batch_dot(b_l_out, w_b_l_pos, (-1, -2)) + K.batch_dot(b_u_out, w_b_l_neg, (-1, -2)) + b_b_l

    return [w_u, b_u, w_l, b_l]



from typing import Any, Dict, Optional, Union

from tensorflow.keras.layers import Layer

import decomon.backward_layers.backward_layers
import decomon.backward_layers.backward_maxpooling
import decomon.backward_layers.backward_merge
import decomon.backward_layers.deel_lip
from decomon.backward_layers.core import BackwardLayer
from decomon.core import BoxDomain, ForwardMode, PerturbationDomain, Slope

_mapping_name2class: Dict[str, Any] = vars(decomon.backward_layers.backward_layers)
_mapping_name2class.update(vars(decomon.backward_layers.deel_lip))
_mapping_name2class.update(vars(decomon.backward_layers.backward_merge))
_mapping_name2class.update(vars(decomon.backward_layers.backward_maxpooling))


def to_backward(
    layer: Layer,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    perturbation_domain: Optional[PerturbationDomain] = None,
    finetune: bool = False,
    **kwargs: Any,
) -> BackwardLayer:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    class_name = layer.__class__.__name__
    if class_name.startswith("Decomon"):
        class_name = "".join(layer.__class__.__name__.split("Decomon")[1:])

    backward_class_name = f"Backward{class_name}"
    try:
        class_ = _mapping_name2class[backward_class_name]
    except KeyError:
        raise NotImplementedError(f"The backward version of {class_name} is not yet implemented.")
    backward_layer_name = f"{layer.name}_backward"
    return class_(
        layer,
        slope=slope,
        mode=mode,
        perturbation_domain=perturbation_domain,
        finetune=finetune,
        dtype=layer.dtype,
        name=backward_layer_name,
        **kwargs,
    )



# creating toy network and assess that the decomposition is correct


import numpy as np
import pytest
import tensorflow.keras.backend as K
from tensorflow.keras.layers import (
    Activation,
    Add,
    Average,
    Conv2D,
    Dense,
    Flatten,
    Input,
)
from tensorflow.keras.models import Model, Sequential

from decomon.core import ForwardMode, get_affine, get_ibp
from decomon.models.forward_cloning import convert_forward


def test_convert_forward_1D(n, mode, floatx, decimal, helpers):
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    input_tensors = helpers.get_input_tensors_for_decomon_convert_from_full_inputs(
        inputs=inputs, mode=mode, dc_decomp=dc_decomp
    )

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)
    inputs_ = helpers.prepare_full_np_inputs_for_convert_model(inputs_, dc_decomp=dc_decomp)
    input_ref_ = helpers.get_input_ref_from_full_inputs(inputs_)

    # keras model and output of reference
    ref_nn = helpers.toy_network_tutorial(dtype=K.floatx())
    output_ref_ = ref_nn.predict(input_ref_)

    # decomon conversion
    _, outputs, _, _ = convert_forward(ref_nn, ibp=ibp, affine=affine, shared=True, input_tensors=input_tensors)

    #  decomon outputs
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # check bounds consistency
    helpers.assert_decomon_model_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )



# Test unit for decomon with Dense layers


import numpy as np
import pytest
import tensorflow.python.keras.backend as K
from numpy.testing import assert_almost_equal
from tensorflow.keras.layers import Conv2D, Input
from tensorflow.keras.models import Model

from decomon.backward_layers.utils_conv import get_toeplitz
from decomon.core import get_affine, get_ibp
from decomon.layers.convert import to_decomon


def test_toeplitz_from_Keras(channels, filter_size, strides, flatten, data_format, padding, floatx, helpers):
    # filter_size, strides, flatten,
    decimal = 5
    if floatx == 16:
        K.set_epsilon(1e-2)
        decimal = 0

    if data_format == "channels_first" and not len(K._get_available_gpus()):
        pytest.skip("data format 'channels first' is possible only in GPU mode")

    dc_decomp = False
    odd, m_0, m_1 = 0, 0, 1

    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)

    # should be working either with convolution of conv2D
    kwargs_layer = dict(
        filters=channels,
        kernel_size=(filter_size, filter_size),
        strides=strides,
        padding=padding,
        use_bias=False,
        dtype=input_ref.dtype,
    )

    layer = Conv2D(**kwargs_layer)

    result_ref = layer(input_ref)
    W = get_toeplitz(layer, flatten)

    if not flatten:
        w_in, h_in, c_in, w_out, h_out, c_out = W.shape
        W = K.reshape(W, (w_in * h_in * c_in, w_out * h_out * c_out))

    n_in, n_out = W.shape
    y_flat = K.reshape(input_ref, (-1, n_in, 1))
    result_flat = K.reshape(result_ref, (-1, n_out))
    result_toeplitz = K.sum(W[None] * y_flat, 1)
    output_test = K.sum((result_toeplitz - result_flat) ** 2)
    f_test = K.function(inputs, output_test)
    output_test_ = f_test(inputs_)

    assert_almost_equal(
        output_test_,
        np.zeros_like(output_test_),
        decimal=decimal,
        err_msg="wrong toeplitz matrix",
    )


def test_toeplitz_from_Decomon(floatx, mode, channels, filter_size, strides, flatten, data_format, padding, helpers):
    if data_format == "channels_first" and not len(K._get_available_gpus()):
        pytest.skip("data format 'channels first' is possible only in GPU mode")

    odd, m_0, m_1 = 0, 0, 1
    decimal = 4
    if floatx == 16:
        K.set_epsilon(1e-2)
        decimal = 0

    dc_decomp = False
    ibp = get_ibp(mode=mode)
    affine = get_affine(mode=mode)

    # tensor inputs
    inputs = helpers.get_tensor_decomposition_images_box(data_format, odd, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_images_box(data_format, odd, m0=m_0, m1=m_1, dc_decomp=dc_decomp)
    kwargs_layer = dict(
        filters=channels,
        kernel_size=(filter_size, filter_size),
        strides=strides,
        padding=padding,
        use_bias=False,
        dtype=input_ref.dtype,
    )

    # should be working either with convolution of conv2D
    layer = Conv2D(**kwargs_layer)
    result_ref = layer(input_ref)

    # toeplitz matrix should be compatible with a DecomonLayer
    input_dim = input_ref.shape[-1]

    decomon_layer = to_decomon(layer, input_dim, ibp=ibp, affine=affine)

    W = get_toeplitz(decomon_layer, flatten)

    if not flatten:
        w_in, h_in, c_in, w_out, h_out, c_out = W.shape
        W = K.reshape(W, (w_in * h_in * c_in, w_out * h_out * c_out))

    n_in, n_out = W.shape
    y_flat = K.reshape(input_ref, (-1, n_in, 1))
    result_flat = K.reshape(result_ref, (-1, n_out))
    result_toeplitz = K.sum(W[None] * y_flat, 1)

    output_test = K.sum((result_toeplitz - result_flat) ** 2)

    f_test = K.function(inputs, output_test)

    output_test_ = f_test(inputs_)

    assert_almost_equal(
        output_test_,
        np.zeros_like(output_test_),
        decimal=decimal,
        err_msg="wrong toeplitz matrix",
    )



# Test unit for decomon with Dense layers


import numpy as np
import pytest
import tensorflow.keras.backend as K
from numpy.testing import assert_almost_equal

from decomon.core import ForwardMode, Slope
from decomon.layers.activations import relu, sigmoid, softmax, softsign, tanh


@pytest.mark.parametrize(
    "activation_func, tensor_func, decimal",
    [
        (sigmoid, K.sigmoid, 5),
        (tanh, K.tanh, 5),
        (softsign, K.softsign, 4),
        (softmax, K.softmax, 4),
        (relu, K.relu, 4),
    ],
)
def test_activation_1D_box(n, mode, floatx, decimal, helpers, activation_func, tensor_func):
    # softmax: test only n=0,3
    if activation_func is softmax:
        if n not in {0, 3}:
            pytest.skip("softmax activation only possible for n=0 or 3")

    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # reference output
    f_ref = K.function(inputs, tensor_func(input_ref))
    output_ref_ = f_ref(inputs_)

    # decomon output
    output = activation_func(inputs_for_mode, dc_decomp=dc_decomp, mode=mode)
    f_decomon = K.function(inputs, output)
    outputs_ = f_decomon(inputs_)

    #  check bounds consistency
    helpers.assert_decomon_layer_output_properties_box(
        full_inputs=inputs_,
        output_ref=output_ref_,
        outputs_for_mode=outputs_,
        mode=mode,
        dc_decomp=dc_decomp,
        decimal=decimal,
    )


@pytest.mark.parametrize(
    "n",
    [
        0,
        1,
        2,
    ],
)
def test_activation_1D_box_slope(n, slope, helpers):
    mode = ForwardMode.AFFINE
    activation_func = relu
    dc_decomp = False

    #  tensor inputs
    inputs = helpers.get_tensor_decomposition_1d_box(dc_decomp=dc_decomp)
    inputs_for_mode = helpers.get_inputs_for_mode_from_full_inputs(inputs=inputs, mode=mode, dc_decomp=dc_decomp)
    input_ref = helpers.get_input_ref_from_full_inputs(inputs=inputs)

    # numpy inputs
    inputs_ = helpers.get_standard_values_1d_box(n, dc_decomp=dc_decomp)

    # decomon output
    outputs = activation_func(inputs_for_mode, dc_decomp=dc_decomp, mode=mode, slope=slope)
    f_decomon = K.function(inputs, outputs)
    outputs_ = f_decomon(inputs_)

    # full outputs & inputs
    (
        z_output,
        u_c_output,
        w_u_output,
        b_u_output,
        l_c_output,
        w_l_output,
        b_l_output,
        h_output,
        g_output,
    ) = helpers.get_full_outputs_from_outputs_for_mode(
        outputs_for_mode=outputs_,
        full_inputs=inputs_,
        mode=mode,
        dc_decomp=dc_decomp,
    )

    (
        x_0,
        y_0,
        z_0,
        u_c_0,
        W_u_0,
        b_u_0,
        l_c_0,
        W_l_0,
        b_l_0,
    ) = inputs_  # numpy values

    # check bounds according to slope
    slope = Slope(slope)
    if n == 0:
        assert_almost_equal(w_u_output, np.zeros(w_u_output.shape))
        assert_almost_equal(b_u_output, np.zeros(b_u_output.shape))
        assert_almost_equal(w_l_output, np.zeros(w_l_output.shape))
        assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
    elif n == 1:
        assert_almost_equal(w_u_output, np.ones(w_u_output.shape))
        assert_almost_equal(b_u_output, np.zeros(b_u_output.shape))
        assert_almost_equal(w_l_output, W_l_0)
        assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
    elif n == 2:
        assert_almost_equal(w_u_output, 0.5 * np.ones(w_u_output.shape))
        assert_almost_equal(b_u_output, 0.5 * np.ones(b_u_output.shape))
        if slope == Slope.O_SLOPE:
            assert_almost_equal(w_l_output, W_l_0)
            assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
        elif slope == Slope.Z_SLOPE:
            assert_almost_equal(w_l_output, np.zeros(w_l_output.shape))
            assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
        elif slope == Slope.V_SLOPE:
            assert_almost_equal(w_l_output, np.zeros(w_l_output.shape))
            assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
        elif slope == Slope.S_SLOPE:
            assert_almost_equal(w_l_output, w_u_output)
            assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
        elif slope == Slope.A_SLOPE:
            assert_almost_equal(b_l_output, np.zeros(b_l_output.shape))
            assert_almost_equal(w_l_output, np.zeros(w_l_output.shape))



from typing import Any, Dict, List, Optional, Tuple, Union

import tensorflow as tf
from tensorflow.keras.layers import Activation, Input, Layer

import decomon.layers.decomon_layers
import decomon.layers.decomon_merge_layers
import decomon.layers.decomon_reshape
import decomon.layers.deel_lip
import decomon.layers.maxpooling
from decomon.core import BoxDomain, ForwardMode, PerturbationDomain, Slope, get_mode
from decomon.layers.core import DecomonLayer

# mapping between decomon class names and actual classes
_mapping_name2class = vars(decomon.layers.decomon_layers)
_mapping_name2class.update(vars(decomon.layers.decomon_merge_layers))
_mapping_name2class.update(vars(decomon.layers.decomon_reshape))
_mapping_name2class.update(vars(decomon.layers.deel_lip))
_mapping_name2class.update(vars(decomon.layers.maxpooling))


def to_decomon(
    layer: Layer,
    input_dim: int,
    slope: Union[str, Slope] = Slope.V_SLOPE,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    finetune: bool = False,
    ibp: bool = True,
    affine: bool = True,
    shared: bool = True,
    fast: bool = True,
) -> DecomonLayer:
    """Transform a standard keras layer into a Decomon layer.

    Type of layer is tested to know how to transform it into a DecomonLayer of the good type.
    If type is not treated yet, raises an TypeError

    Args:
        layer: a Keras Layer
        input_dim: an integer that represents the dim
            of the input perturbation domain
        slope:
        dc_decomp: boolean that indicates whether we return a difference
            of convex decomposition of our layer
        perturbation_domain: the type of perturbation domain
        ibp: boolean that indicates whether we propagate constant bounds
        affine: boolean that indicates whether we propagate affine
            bounds

    Returns:
        the associated DecomonLayer
    """

    # get class name
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    mode = get_mode(ibp=ibp, affine=affine)
    layer_decomon = _to_decomon_wo_input_init(
        layer=layer,
        namespace=_mapping_name2class,
        slope=slope,
        dc_decomp=dc_decomp,
        perturbation_domain=perturbation_domain,
        finetune=finetune,
        mode=mode,
        shared=shared,
        fast=fast,
    )

    input_tensors = _prepare_input_tensors(
        layer=layer, input_dim=input_dim, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode
    )
    layer_decomon(input_tensors)
    layer_decomon.reset_layer(layer)

    # return layer_decomon
    return layer_decomon


def _to_decomon_wo_input_init(
    layer: Layer,
    namespace: Dict[str, Any],
    slope: Union[str, Slope] = Slope.V_SLOPE,
    dc_decomp: bool = False,
    perturbation_domain: Optional[PerturbationDomain] = None,
    finetune: bool = False,
    mode: ForwardMode = ForwardMode.HYBRID,
    shared: bool = True,
    fast: bool = True,
) -> DecomonLayer:
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    class_name = layer.__class__.__name__
    # check if layer has a built argument that built is set to True
    if hasattr(layer, "built"):
        if not layer.built:
            raise ValueError(f"the layer {layer.name} has not been built yet")
    decomon_class_name = f"Decomon{class_name}"
    config_layer = layer.get_config()
    config_layer["name"] = layer.name + "_decomon"
    config_layer["dc_decomp"] = dc_decomp
    config_layer["perturbation_domain"] = perturbation_domain

    config_layer["mode"] = mode
    config_layer["finetune"] = finetune
    config_layer["slope"] = slope
    config_layer["shared"] = shared
    config_layer["fast"] = fast
    if not isinstance(layer, Activation):
        config_layer.pop("activation", None)  # Hyp: no non-linear activation in dense or conv2d layers
    try:
        layer_decomon = namespace[decomon_class_name].from_config(config_layer)
    except:
        raise NotImplementedError(f"The decomon version of {class_name} is not yet implemented.")

    layer_decomon.share_weights(layer)
    return layer_decomon


def _prepare_input_tensors(
    layer: Layer, input_dim: int, dc_decomp: bool, perturbation_domain: PerturbationDomain, mode: ForwardMode
) -> List[tf.Tensor]:
    original_input_shapes = get_layer_input_shape(layer)
    decomon_input_shapes: List[List[Optional[int]]] = [list(input_shape[1:]) for input_shape in original_input_shapes]
    n_input = len(decomon_input_shapes)
    x_input_shape = perturbation_domain.get_x_input_shape(input_dim)
    x_input = Input(x_input_shape, dtype=layer.dtype)
    w_input = [Input(tuple([input_dim] + decomon_input_shapes[i])) for i in range(n_input)]
    y_input = [Input(tuple(decomon_input_shapes[i])) for i in range(n_input)]

    if mode == ForwardMode.HYBRID:
        nested_input_list = [
            [x_input, y_input[i], w_input[i], y_input[i], y_input[i], w_input[i], y_input[i]] for i in range(n_input)
        ]
    elif mode == ForwardMode.IBP:
        nested_input_list = [[y_input[i], y_input[i]] for i in range(n_input)]
    elif mode == ForwardMode.AFFINE:
        nested_input_list = [[x_input, w_input[i], y_input[i], w_input[i], y_input[i]] for i in range(n_input)]
    else:
        raise ValueError(f"Unknown mode {mode}")

    flatten_input_list = [tensor for input_list in nested_input_list for tensor in input_list]

    if dc_decomp:
        if n_input == 1:
            flatten_input_list += [y_input[0], y_input[0]]
        else:
            raise NotImplementedError()

    return flatten_input_list


SingleInputShapeType = Tuple[Optional[int]]


def get_input_shapes_as_a_list_of_input_shape(
    input_shapes: Union[List[SingleInputShapeType], SingleInputShapeType]
) -> List[SingleInputShapeType]:
    if not isinstance(input_shapes, list):
        return [input_shapes]
    else:
        return input_shapes


def get_layer_input_shape(layer: Layer) -> List[SingleInputShapeType]:
    """Retrieves the input shape(s) of a layer.

    Only applicable if the layer has exactly one input,
    i.e. if it is connected to one incoming layer, or if all inputs
    have the same shape.

    Args:
        layer:

    Returns:
        Input shape, as an integer shape tuple
        (or list of shape tuples, one tuple per input tensor).

    Raises:
        AttributeError: if the layer has no defined input_shape.
        RuntimeError: if called in Eager mode.
    """

    if not layer.inbound_nodes:
        raise AttributeError(
            f'The layer "{layer.name}" has never been called '
            "and thus has no defined input shape. Note that the "
            "`input_shape` property is only available for "
            "Functional and Sequential models."
        )
    all_input_shapes = set(
        [str(get_input_shapes_as_a_list_of_input_shape(node.input_shapes)) for node in layer.inbound_nodes]
    )
    if len(all_input_shapes) == 1:
        return get_input_shapes_as_a_list_of_input_shape(layer.inbound_nodes[0].input_shapes)
    else:
        raise AttributeError(
            'The layer "' + str(layer.name) + '" has multiple inbound nodes, '
            "with different input shapes. Hence "
            'the notion of "input shape" is '
            "ill-defined for the layer. "
            "Use `get_input_shape_at(node_index)` "
            "instead."
        )



from typing import Any, List, Optional, Union

import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Flatten

from decomon.core import (
    BoxDomain,
    ForwardMode,
    GridDomain,
    InputsOutputsSpec,
    PerturbationDomain,
    Slope,
    get_affine,
    get_ibp,
)
from decomon.layers.utils import sort
from decomon.utils import get_linear_hull_relu, maximum, minus, relu_, subtract


def backward_add(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
) -> List[List[tf.Tensor]]:
    """Backward  LiRPA of inputs_0+inputs_1

    Args:
        inputs_0
        inputs_1
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    op_flat = Flatten(dtype=K.floatx())  # pas terrible  a revoir
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x_0, u_c_0, w_u_0, b_u_0, l_c_0, w_l_0, b_l_0, h_0, g_0 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_0
    )
    x_1, u_c_1, w_u_1, b_u_1, l_c_1, w_l_1, b_l_1, h_1, g_1 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_1
    )

    u_c_0 = op_flat(u_c_0)
    u_c_1 = op_flat(u_c_1)
    l_c_0 = op_flat(l_c_0)
    l_c_1 = op_flat(l_c_1)

    x_0 = K.concatenate([K.expand_dims(l_c_0, 1), K.expand_dims(u_c_0, 1)], 1)
    x_1 = K.concatenate([K.expand_dims(l_c_1, 1), K.expand_dims(u_c_1, 1)], 1)
    upper_0 = BoxDomain().get_upper(x_0, w_u_out, b_u_out)
    upper_1 = BoxDomain().get_upper(x_1, w_u_out, b_u_out)
    lower_0 = BoxDomain().get_lower(x_0, w_l_out, b_l_out)
    lower_1 = BoxDomain().get_lower(x_1, w_l_out, b_l_out)

    w_u_out_0 = w_u_out
    b_u_out_0 = upper_1
    w_l_out_0 = w_l_out
    b_l_out_0 = lower_1

    w_u_out_1 = w_u_out
    b_u_out_1 = upper_0
    w_l_out_1 = w_l_out
    b_l_out_1 = lower_0

    return [[w_u_out_0, b_u_out_0, w_l_out_0, b_l_out_0], [w_u_out_1, b_u_out_1, w_l_out_1, b_l_out_1]]


def backward_linear_prod(
    x_0: tf.Tensor,
    bounds_x: List[tf.Tensor],
    back_bounds: List[tf.Tensor],
    perturbation_domain: Optional[PerturbationDomain] = None,
) -> List[tf.Tensor]:
    """Backward  LiRPA of a subroutine prod

    Args:
        bounds_x
        back_bounds

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()

    z_value = K.cast(0.0, x_0.dtype)
    o_value = K.cast(1.0, x_0.dtype)

    w_u_i, b_u_i, w_l_i, b_l_i = bounds_x
    w_u, b_u, w_l, b_l = back_bounds

    if len(w_u_i.shape) > 3:
        n_dim = w_u_i.get_input_shape_at(0)[1]
        w_u_i = K.reshape(w_u_i, (-1, n_dim, n_dim))
        w_l_i = K.reshape(w_l_i, (-1, n_dim, n_dim))
        b_u_i = K.reshape(b_u_i, (-1, n_dim))
        b_l_i = K.reshape(b_l_i, (-1, n_dim))

    x_max = perturbation_domain.get_upper(x_0, w_u_i - w_l_i, b_u_i - b_l_i)
    mask_b = o_value - K.sign(x_max)
    mask_a = o_value - mask_b

    w_u_i_expanded = K.expand_dims(K.expand_dims(w_u_i, 1), -1)
    w_l_i_expanded = K.expand_dims(K.expand_dims(w_l_i, 1), -1)
    b_u_i_expanded = K.expand_dims(K.expand_dims(b_u_i, 1), -1)
    b_l_i_expanded = K.expand_dims(K.expand_dims(b_l_i, 1), -1)
    mask_a = K.expand_dims(K.expand_dims(mask_a, 1), -1)
    mask_b = K.expand_dims(K.expand_dims(mask_b, 1), -1)

    w_u_pos = K.maximum(w_u, z_value)
    w_u_neg = K.minimum(w_u, z_value)
    w_l_pos = K.maximum(w_l, z_value)
    w_l_neg = K.minimum(w_l, z_value)

    w_u_pos_expanded = K.expand_dims(w_u_pos, 2)
    w_u_neg_expanded = K.expand_dims(w_u_neg, 2)
    w_l_pos_expanded = K.expand_dims(w_l_pos, 2)
    w_l_neg_expanded = K.expand_dims(w_l_neg, 2)
    mask_a_expanded = K.expand_dims(mask_a, 2)
    mask_b_expanded = K.expand_dims(mask_b, 2)

    w_u_out = K.sum(
        mask_a_expanded * (w_u_pos_expanded * w_u_i_expanded + w_u_neg_expanded * w_l_i_expanded), 3
    ) + K.sum(K.expand_dims(w_u, 2) * mask_b_expanded * w_u_i_expanded, 3)
    w_l_out = K.sum(
        mask_a_expanded * (w_l_pos_expanded * w_l_i_expanded + w_l_neg_expanded * w_u_i_expanded), 3
    ) + K.sum(K.expand_dims(w_l, 2) * mask_b_expanded * w_l_i_expanded, 3)

    b_u_out = (
        K.sum(mask_a * (w_u_pos * b_u_i_expanded + w_u_neg * b_l_i_expanded), 2)
        + K.sum(mask_b * (w_u * b_u_i_expanded), 2)
        + b_u
    )
    b_l_out = (
        K.sum(mask_a * (w_l_pos * b_l_i_expanded + w_l_neg * b_u_i_expanded), 2)
        + K.sum(mask_b * (w_l * b_l_i_expanded), 2)
        + b_l
    )

    return [w_u_out, b_u_out, w_l_out, b_l_out]


def backward_maximum(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
    **kwargs: Any,
) -> List[List[tf.Tensor]]:
    """Backward  LiRPA of maximum(inputs_0, inputs_1)

    Args:
        inputs_0
        inputs_1
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    input_step_a_0 = subtract(
        inputs_0, inputs_1, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode
    )
    input_step_0 = relu_(
        input_step_a_0, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, **kwargs
    )
    _, bounds_1 = backward_add(
        input_step_0,
        inputs_1,
        w_u_out,
        b_u_out,
        w_l_out,
        b_l_out,
        perturbation_domain=perturbation_domain,
        mode=mode,
        dc_decomp=dc_decomp,
    )

    input_step_a_1 = subtract(
        inputs_1, inputs_0, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode
    )
    input_step_1 = relu_(
        input_step_a_1, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain, mode=mode, **kwargs
    )
    _, bounds_0 = backward_add(
        input_step_1,
        inputs_0,
        w_u_out,
        b_u_out,
        w_l_out,
        b_l_out,
        perturbation_domain=perturbation_domain,
        mode=mode,
        dc_decomp=dc_decomp,
    )

    return [bounds_0, bounds_1]


# convex hull of the maximum between two functions
def backward_max_(
    inputs: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    axis: int = -1,
    dc_decomp: bool = False,
    **kwargs: Any,
) -> List[tf.Tensor]:
    """Backward  LiRPA of max

    Args:
        inputs: list of tensors
        dc_decomp: boolean that indicates
        grad_bounds: boolean that indicates whether
        perturbation_domain: the type of perturbation domain
        axis: axis to perform the maximum
    whether we return a difference of convex decomposition of our layer
    we propagate upper and lower bounds on the values of the gradient

    Returns:
        max operation  along an axis
    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    ibp = get_ibp(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs, tight=False, compute_ibp_from_affine=False
    )
    dtype = x.dtype
    input_shape = inputs_outputs_spec.get_input_shape(inputs)
    max_dim = input_shape[axis]
    empty_tensor = inputs_outputs_spec.get_empty_tensor(dtype=dtype)
    empty_tensor_list = [empty_tensor] * max_dim
    z_value = K.cast(0.0, dtype=dtype)

    # do some transpose so that the last axis is also at the end
    if ibp:
        u_c_list = tf.split(u_c, max_dim, axis)
        l_c_list = tf.split(l_c, max_dim, axis)
        u_c_tmp = u_c_list[0] + z_value * (u_c_list[0])
        l_c_tmp = l_c_list[0] + z_value * (l_c_list[0])
    else:
        u_c_list, l_c_list = empty_tensor_list, empty_tensor_list
        u_c_tmp, l_c_tmp = empty_tensor, empty_tensor

    if affine:
        b_u_list = tf.split(b_u, max_dim, axis)
        b_l_list = tf.split(b_l, max_dim, axis)
        b_u_tmp = b_u_list[0] + z_value * (b_u_list[0])
        b_l_tmp = b_l_list[0] + z_value * (b_l_list[0])

        if axis == -1:
            w_u_list = tf.split(w_u, max_dim, axis)
            w_l_list = tf.split(w_l, max_dim, axis)
        else:
            w_u_list = tf.split(w_u, max_dim, axis + 1)
            w_l_list = tf.split(w_l, max_dim, axis + 1)
        w_u_tmp = w_u_list[0] + z_value * (w_u_list[0])
        w_l_tmp = w_l_list[0] + z_value * (w_l_list[0])
    else:
        b_u_list, b_l_list, w_u_list, w_l_list = (
            empty_tensor_list,
            empty_tensor_list,
            empty_tensor_list,
            empty_tensor_list,
        )
        b_u_tmp, b_l_tmp, w_u_tmp, w_l_tmp = empty_tensor, empty_tensor, empty_tensor, empty_tensor

    if dc_decomp:
        raise NotImplementedError()
    else:
        h_list, g_list = empty_tensor_list, empty_tensor_list
        h_tmp, g_tmp = empty_tensor, empty_tensor

    outputs = []
    output_tmp = inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
        [x, u_c_tmp, w_u_tmp, b_u_tmp, l_c_tmp, w_l_tmp, b_l_tmp, h_tmp, g_tmp]
    )
    for i in range(1, max_dim):
        output_i = inputs_outputs_spec.extract_outputsformode_from_fulloutputs(
            [x, u_c_list[i], w_u_list[i], b_u_list[i], l_c_list[i], w_l_list[i], b_l_list[i], h_list[i], g_list[i]]
        )
        outputs.append([[elem for elem in output_tmp], output_i])
        output_tmp = maximum(
            output_tmp, output_i, dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain
        )

    outputs = outputs[::-1]
    bounds = []
    if len(outputs) > 0:
        for input_0, input_1 in outputs:
            bounds_0, bounds_1 = backward_maximum(
                input_0, input_1, w_u_out, b_u_out, w_l_out, b_l_out, mode=mode, dc_decomp=dc_decomp, **kwargs
            )
            bounds.append(bounds_1)
            w_u_out, b_u_out, w_l_out, b_l_out = bounds_0
        bounds.append(bounds_0)
        bounds = bounds[::-1]

    if axis < 0:
        w_u_out = K.concatenate([b[0] for b in bounds], axis - 1)
        w_l_out = K.concatenate([b[2] for b in bounds], axis - 1)
        b_u_out = K.sum(K.concatenate([K.expand_dims(b[1], axis - 1) for b in bounds], axis - 1), axis - 1)
        b_l_out = K.sum(K.concatenate([K.expand_dims(b[3], axis - 1) for b in bounds], axis - 1), axis - 1)
    else:
        w_u_out = K.concatenate([b[0] for b in bounds], axis)
        w_l_out = K.concatenate([b[2] for b in bounds], axis)
        b_u_out = K.sum(K.concatenate([K.expand_dims(b[1], axis) for b in bounds], axis), axis)
        b_l_out = K.sum(K.concatenate([K.expand_dims(b[3], axis) for b in bounds], axis), axis)

    return [w_u_out, b_u_out, w_l_out, b_l_out]


def backward_minimum(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
    **kwargs: Any,
) -> List[List[tf.Tensor]]:
    """Backward  LiRPA of minimum(inputs_0, inputs_1)

    Args:
        inputs_0
        inputs_1
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    w_u_out, b_u_out, w_l_out, b_l_out = backward_minus(w_u_out, b_u_out, w_l_out, b_l_out)
    bounds_0, bounds_1 = backward_maximum(
        inputs_0,
        inputs_1,
        w_u_out,
        b_u_out,
        w_l_out,
        b_l_out,
        perturbation_domain=perturbation_domain,
        mode=mode,
        dc_decomp=dc_decomp,
        **kwargs,
    )

    w_u_out_0, b_u_out_0, w_l_out_0, b_l_out_0 = bounds_0
    w_u_out_1, b_u_out_1, w_l_out_1, b_l_out_1 = bounds_1

    bounds_0 = backward_minus(w_u_out_0, b_u_out_0, w_l_out_0, b_l_out_0)
    bounds_1 = backward_minus(w_u_out_1, b_u_out_1, w_l_out_1, b_l_out_1)

    return [bounds_0, bounds_1]


def backward_minus(
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
) -> List[tf.Tensor]:
    """Backward  LiRPA of -x

    Args:
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        mode

    Returns:

    """
    return [-w_l_out, -b_l_out, -w_u_out, -b_u_out]


def backward_scale(
    scale_factor: float,
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
) -> List[tf.Tensor]:
    """Backward  LiRPA of scale_factor*x

    Args:
        scale_factor
        w_u_out
        b_u_out
        w_l_out
        b_l_out

    Returns:

    """

    if scale_factor >= 0:
        output = [scale_factor * w_u_out, b_u_out, scale_factor * w_l_out, b_l_out]
    else:
        output = [scale_factor * w_l_out, b_l_out, scale_factor * w_u_out, b_u_out]

    return output


def backward_subtract(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
) -> List[List[tf.Tensor]]:
    """Backward  LiRPA of inputs_0 - inputs_1

    Args:
        inputs_0
        inputs_1
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    inputs_1 = minus(inputs_1, mode=mode, dc_decomp=dc_decomp, perturbation_domain=perturbation_domain)
    bounds_0, bounds_1 = backward_add(
        inputs_0,
        inputs_1,
        w_u_out,
        b_u_out,
        w_l_out,
        b_l_out,
        perturbation_domain=perturbation_domain,
        mode=mode,
        dc_decomp=dc_decomp,
    )

    bounds_1 = [-bounds_1[0], bounds_1[1], -bounds_1[2], bounds_1[3]]
    return [bounds_0, bounds_1]


def backward_multiply(
    inputs_0: List[tf.Tensor],
    inputs_1: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
) -> List[List[tf.Tensor]]:
    """Backward  LiRPA of element-wise multiply inputs_0*inputs_1

    Args:
        inputs_0
        inputs_1
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        perturbation_domain
        mode

    Returns:

    """

    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x_0, u_c_0, w_u_0, b_u_0, l_c_0, w_l_0, b_l_0, h_0, g_0 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_0
    )
    x_1, u_c_1, w_u_1, b_u_1, l_c_1, w_l_1, b_l_1, h_1, g_1 = inputs_outputs_spec.get_fullinputs_from_inputsformode(
        inputs_1
    )

    z_value = K.cast(0.0, u_c_0.dtype)

    n = np.prod(u_c_0.shape[1:])
    n_shape = [-1, n]
    # broadcast dimensions if needed
    n_out = len(w_u_out.shape[1:])
    for _ in range(n_out):
        n_shape += [1]
    a_u_0 = K.reshape(u_c_1, n_shape)
    a_u_1 = K.reshape(u_c_0, n_shape)

    b_u_0 = K.reshape((K.maximum(l_c_0, z_value) * u_c_1 + K.minimum(l_c_0, z_value) * l_c_1 - u_c_1 * l_c_0), n_shape)
    b_u_1 = K.reshape((K.maximum(l_c_1, z_value) * u_c_0 + K.minimum(l_c_1, z_value) * l_c_0 - u_c_0 * l_c_1), n_shape)

    a_l_0 = K.reshape(l_c_1, n_shape)
    a_l_1 = K.reshape(l_c_0, n_shape)
    b_l_0 = K.reshape((K.maximum(l_c_0, z_value) * l_c_1 + K.minimum(l_c_0, z_value) * u_c_1 - l_c_1 * l_c_0), n_shape)
    b_l_1 = K.reshape((K.maximum(l_c_1, z_value) * l_c_0 + K.minimum(l_c_1, z_value) * u_c_0 - l_c_0 * l_c_1), n_shape)

    # upper
    w_u_out_max = K.maximum(w_u_out, z_value)
    w_u_out_min = K.minimum(w_u_out, z_value)

    w_u_out_0 = w_u_out_max * a_u_0 + w_u_out_min * a_l_0
    w_u_out_1 = w_u_out_max * a_u_1 + w_u_out_min * a_l_1

    b_u_out_0 = K.sum(w_u_out_max * b_u_0, 1) + K.sum(w_u_out_min * b_l_0, 1) + b_u_out
    b_u_out_1 = K.sum(w_u_out_max * b_u_1, 1) + K.sum(w_u_out_min * b_l_1, 1) + b_u_out

    # lower
    w_l_out_max = K.maximum(w_l_out, z_value)
    w_l_out_min = K.minimum(w_l_out, z_value)

    w_l_out_0 = w_l_out_max * a_l_0 + w_l_out_min * a_u_0
    w_l_out_1 = w_l_out_max * a_l_1 + w_l_out_min * a_u_1

    b_l_out_0 = K.sum(w_l_out_max * b_l_0, 1) + K.sum(w_l_out_min * b_u_0, 1) + b_l_out
    b_l_out_1 = K.sum(w_l_out_max * b_l_1, 1) + K.sum(w_l_out_min * b_u_1, 1) + b_l_out

    bounds_0 = [w_u_out_0, b_u_out_0, w_l_out_0, b_l_out_0]
    bounds_1 = [w_u_out_1, b_u_out_1, w_l_out_1, b_l_out_1]

    return [bounds_0, bounds_1]


def backward_sort(
    inputs: List[tf.Tensor],
    w_u_out: tf.Tensor,
    b_u_out: tf.Tensor,
    w_l_out: tf.Tensor,
    b_l_out: tf.Tensor,
    axis: int = -1,
    perturbation_domain: Optional[PerturbationDomain] = None,
    mode: Union[str, ForwardMode] = ForwardMode.HYBRID,
    dc_decomp: bool = False,
) -> List[tf.Tensor]:
    """Backward  LiRPA of sort

    Args:
        inputs
        w_u_out
        b_u_out
        w_l_out
        b_l_out
        axis
        perturbation_domain
        mode

    Returns:

    """
    if perturbation_domain is None:
        perturbation_domain = BoxDomain()
    mode = ForwardMode(mode)
    affine = get_affine(mode)
    z_value = K.cast(0.0, w_u_out.dtype)
    inputs_outputs_spec = InputsOutputsSpec(dc_decomp=dc_decomp, mode=mode, perturbation_domain=perturbation_domain)
    x, u_c, w_u, b_u, l_c, w_l, b_l, h, g = inputs_outputs_spec.get_fullinputs_from_inputsformode(inputs)

    # build fake inputs with no linearity
    n_dim = np.prod(u_c.shape[1:])
    w_tmp = z_value * K.concatenate([u_c[:, None] * n_dim], 1)

    inputs_tmp = [x, u_c, w_tmp, u_c, l_c, w_tmp, l_c]
    outputs_tmp = sort(
        inputs_tmp, axis=axis, perturbation_domain=perturbation_domain, mode=ForwardMode.HYBRID, dc_decomp=False
    )
    _, _, w_u_tmp, b_u_tmp, _, w_l_tmp, b_l_tmp = outputs_tmp

    # w_u_tmp (None, n_dim, y.shape[1:)
    # w_u_out (None, 1, n_dim, n_out)
    w_u_tmp = K.reshape(w_u_tmp, [-1, n_dim, n_dim])  # (None, n_dim, n_dim)
    b_u_tmp = K.reshape(b_u_tmp, [-1, n_dim])  # (None, n_dim)
    w_l_tmp = K.reshape(w_l_tmp, [-1, n_dim, n_dim])
    b_l_tmp = K.reshape(b_l_tmp, [-1, n_dim])

    # combine with backward bounds
    w_u_out_pos = K.maximum(w_u_out, z_value)  # (None, 1, n_dim, n_out)
    w_u_out_neg = K.minimum(w_u_out, z_value)
    w_l_out_pos = K.maximum(w_l_out, z_value)
    w_l_out_neg = K.minimum(w_l_out, z_value)

    w_u_out = K.sum(w_u_out_pos * K.expand_dims(w_u_tmp, -1) + w_u_out_pos * K.expand_dims(w_l_tmp, -1), 1)
    w_l_out = K.sum(w_u_out_pos * K.expand_dims(w_u_tmp, -1) + w_u_out_pos * K.expand_dims(w_l_tmp, -1), 1)
    b_u_out = b_u_out + K.sum(w_u_out_pos * b_u_tmp, 1) + K.sum(w_u_out_neg * b_l_tmp, 1)
    b_l_out = b_l_out + K.sum(w_l_out_pos * b_l_tmp, 1) + K.sum(w_l_out_neg * b_u_tmp, 1)

    return [w_u_out, b_u_out, w_l_out, b_l_out]


def get_identity_lirpa(inputs: List[tf.Tensor]) -> List[tf.Tensor]:
    y = inputs[-1]
    shape = np.prod(y.shape[1:])

    z_value = K.cast(0.0, y.dtype)
    o_value = K.cast(1.0, y.dtype)
    y_flat = K.reshape(y, [-1, shape])

    w_u_out, w_l_out = [o_value + z_value * y_flat] * 2
    b_u_out, b_l_out = [z_value * y_flat] * 2
    w_u_out = tf.linalg.diag(w_u_out)
    w_l_out = tf.linalg.diag(w_l_out)

    return [w_u_out, b_u_out, w_l_out, b_l_out]



import copy
from typing import Any, Dict, List, Optional, Union

import tensorflow as tf

from decomon.core import (
    BoxDomain,
    GridDomain,
    InputsOutputsSpec,
    Option,
    PerturbationDomain,
)
from decomon.models.utils import ConvertMethod

try:
    from keras.src.engine.functional import (
        get_network_config,  # new path starting from keras 2.13
    )
except ImportError:
    from keras.engine.functional import get_network_config  # old path until keras 2.12


class DecomonModel(tf.keras.Model):
    def __init__(
        self,
        inputs: Union[tf.Tensor, List[tf.Tensor]],
        outputs: Union[tf.Tensor, List[tf.Tensor]],
        perturbation_domain: Optional[PerturbationDomain] = None,
        dc_decomp: bool = False,
        method: Union[str, ConvertMethod] = ConvertMethod.FORWARD_AFFINE,
        ibp: bool = True,
        affine: bool = True,
        finetune: bool = False,
        shared: bool = True,
        backward_bounds: bool = False,
        **kwargs: Any,
    ):
        super().__init__(inputs, outputs, **kwargs)
        if perturbation_domain is None:
            perturbation_domain = BoxDomain()
        self.perturbation_domain = perturbation_domain
        self.dc_decomp = dc_decomp
        self.method = ConvertMethod(method)
        self.ibp = ibp
        self.affine = affine
        self.finetune = finetune
        self.backward_bounds = backward_bounds
        self.shared = shared

    def set_domain(self, perturbation_domain: PerturbationDomain) -> None:
        perturbation_domain = _check_domain(self.perturbation_domain, perturbation_domain)
        self.perturbation_domain = perturbation_domain
        for layer in self.layers:
            if hasattr(layer, "perturbation_domain"):
                layer.perturbation_domain = self.perturbation_domain

    def freeze_weights(self) -> None:
        for layer in self.layers:
            if hasattr(layer, "freeze_weights"):
                layer.freeze_weights()

    def unfreeze_weights(self) -> None:
        for layer in self.layers:
            if hasattr(layer, "unfreeze_weights"):
                layer.unfreeze_weights()

    def freeze_alpha(self) -> None:
        for layer in self.layers:
            if hasattr(layer, "freeze_alpha"):
                layer.freeze_alpha()

    def unfreeze_alpha(self) -> None:
        for layer in self.layers:
            if hasattr(layer, "unfreeze_alpha"):
                layer.unfreeze_alpha()

    def reset_finetuning(self) -> None:
        for layer in self.layers:
            if hasattr(layer, "reset_finetuning"):
                layer.reset_finetuning()

    def get_config(self) -> Dict[str, Any]:
        # Continue adding configs into what the super class has added.
        config = super().get_config()
        return copy.deepcopy(get_network_config(self, config=config))


def _check_domain(
    perturbation_domain_prev: PerturbationDomain, perturbation_domain: PerturbationDomain
) -> PerturbationDomain:
    if type(perturbation_domain) != type(perturbation_domain_prev):
        raise NotImplementedError("We can only change the parameters of the perturbation domain, not its type.")

    return perturbation_domain


def get_AB(model: DecomonModel) -> Dict[str, List[tf.Variable]]:
    dico_AB: Dict[str, List[tf.Variable]] = {}
    perturbation_domain = model.perturbation_domain
    if not (isinstance(perturbation_domain, GridDomain) and perturbation_domain.opt_option == Option.milp):
        return dico_AB

    for layer in model.layers:
        name = layer.name
        sub_names = name.split("backward_activation")
        if len(sub_names) > 1:
            key = f"{layer.layer.name}_{layer.rec}"
            if key not in dico_AB:
                dico_AB[key] = layer.grid_finetune
    return dico_AB


def get_AB_finetune(model: DecomonModel) -> Dict[str, tf.Variable]:
    dico_AB: Dict[str, tf.Variable] = {}
    perturbation_domain = model.perturbation_domain
    if not (isinstance(perturbation_domain, GridDomain) and perturbation_domain.opt_option == Option.milp):
        return dico_AB

    if not model.finetune:
        return dico_AB

    for layer in model.layers:
        name = layer.name
        sub_names = name.split("backward_activation")
        if len(sub_names) > 1:
            key = f"{layer.layer.name}_{layer.rec}"
            if key not in dico_AB:
                dico_AB[key] = layer.alpha_b_l
    return dico_AB



from decomon.core import BoxDomain, ForwardMode
from decomon.models.crown import Convert2BackwardMode, Fuse, MergeWithPrevious


def test_crown_layers():
    mode = ForwardMode.AFFINE
    perturbation_domain = BoxDomain()
    input_shape_layer = (1, 2, 4)
    backward_shape_layer = (2, 5, 10)

    layer = Fuse(mode=mode)
    config = layer.get_config()
    assert config["mode"] == mode

    layer = Convert2BackwardMode(mode=mode, perturbation_domain=perturbation_domain)
    config = layer.get_config()
    assert config["mode"] == mode
    assert "perturbation_domain" in config

    layer = MergeWithPrevious(input_shape_layer=input_shape_layer, backward_shape_layer=backward_shape_layer)
    config = layer.get_config()
    assert config["input_shape_layer"] == input_shape_layer
    assert config["backward_shape_layer"] == backward_shape_layer



from typing import Callable, List, Optional, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt
import tensorflow as tf

from decomon.core import BallDomain, BoxDomain, GridDomain
from decomon.models.convert import clone
from decomon.models.models import DecomonModel
from decomon.models.utils import ConvertMethod

IntegerType = Union[int, np.int_]
"""Alias for integers types."""
LabelType = Union[IntegerType, Sequence[IntegerType], np.typing.NDArray[np.int_]]
"""Alias for labels types."""


def _get_dual_ord(p: float) -> float:
    if p == np.inf:
        return 1
    elif p == 1:
        return np.inf
    elif p == 2:
        return 2
    else:
        raise ValueError(f"p must be equal to 1, 2, or np.inf, unknown value {p}.")


def prepare_labels(labels: LabelType, n_batch: int) -> np.typing.NDArray[np.int_]:
    if isinstance(labels, (int, np.int_)):
        labels = np.zeros((n_batch, 1), dtype=np.int_) + labels
    elif not isinstance(labels, np.ndarray):
        labels = np.array(labels).reshape((n_batch, -1))
    else:
        labels = labels.reshape((n_batch, -1))
    return labels.astype("int_")


##### ADVERSARIAL ROBUSTTNESS #####
def get_adv_box(
    model: Union[tf.keras.Model, DecomonModel],
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    source_labels: LabelType,
    target_labels: Optional[LabelType] = None,
    batch_size: int = -1,
    n_sub_boxes: int = 1,
) -> npt.NDArray[np.float_]:
    """if the output is negative, then it is a formal guarantee that there is no adversarial examples

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        source_labels: the list of label that should be predicted all
            the time in the box (either an integer, either an array that
            can contain multiple source labels for each sample)
        target_labels: the list of label that should never be predicted
            in the box (either an integer, either an array that can
            contain multiple target labels for each sample)
        batch_size: for computational efficiency, one can split the
            calls to minibatches
        n_sub_boxes:
    Returns:
        numpy array, vector with upper bounds for adversarial attacks
    """
    if np.min(x_max - x_min) < 0:
        raise UserWarning("Inconsistency Error: x_max < x_min")

    # check that the model is a DecomonModel, else do the conversion
    decomon_model: DecomonModel
    if not isinstance(model, DecomonModel):
        decomon_model = clone(model)
    else:
        assert isinstance(model.perturbation_domain, (BoxDomain, GridDomain))
        decomon_model = model

    n_split = 1
    n_batch = len(x_min)

    if n_sub_boxes > 1:
        x_min, x_max = refine_boxes(x_min, x_max, n_sub_boxes)
        # reshape
        n_split = x_min.shape[1]
        shape = list(x_min.shape[2:])
        x_min = np.reshape(x_min, [-1] + shape)
        x_max = np.reshape(x_max, [-1] + shape)

    # reshape x_mmin, x_max
    if decomon_model.backward_bounds:
        input_shape = list(decomon_model.input_shape[0][2:])
    else:
        input_shape = list(decomon_model.input_shape[2:])
    input_dim = np.prod(input_shape)
    x_reshaped = x_min + 0 * x_min
    x_reshaped = x_reshaped.reshape([-1] + input_shape)

    x_min = x_min.reshape((-1, 1, input_dim))
    x_max = x_max.reshape((-1, 1, input_dim))

    z = np.concatenate([x_min, x_max], 1)

    source_labels = prepare_labels(source_labels, n_batch)
    if target_labels is not None:
        target_labels = prepare_labels(target_labels, n_batch)

    if n_split > 1:
        shape = list(source_labels.shape[1:])
        source_labels = np.reshape(np.concatenate([source_labels[:, None]] * n_split, 1), [-1] + shape)
        if target_labels is not None:
            target_labels = np.reshape(np.concatenate([target_labels[:, None]] * n_split, 1), [-1] + shape)

    if batch_size > 0:
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_min_list = [x_min[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        x_max_list = [x_max[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        source_labels_list = [
            source_labels[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)
        ]
        if target_labels is not None:
            target_labels_list = [
                target_labels[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)
            ]
            adv_score = np.concatenate(
                [
                    get_adv_box(
                        decomon_model, x_min_list[i], x_max_list[i], source_labels_list[i], target_labels_list[i], -1
                    )
                    for i in range(len(x_min_list))
                ]
            )
        else:
            adv_score = np.concatenate(
                [
                    get_adv_box(decomon_model, x_min_list[i], x_max_list[i], source_labels_list[i], None, -1)
                    for i in range(len(x_min_list))
                ]
            )

    else:
        ibp = decomon_model.ibp
        affine = decomon_model.affine
        n_label = source_labels.shape[-1]

        # two possitible cases: the model improves the bound based on the knowledge of the labels
        output: npt.NDArray[np.float_]
        if decomon_model.backward_bounds:
            C = np.diag([1] * n_label)[None] - source_labels[:, :, None]
            output = decomon_model.predict([z, C], verbose=0)
        else:
            output = decomon_model.predict(z, verbose=0)

        def get_ibp_score(
            u_c: npt.NDArray[np.float_],
            l_c: npt.NDArray[np.float_],
            source_tensor: npt.NDArray[np.int_],
            target_tensor: Optional[npt.NDArray[np.int_]] = None,
            backward: bool = False,
        ) -> npt.NDArray[np.float_]:
            if target_tensor is None:
                target_tensor = 1 - source_tensor

            shape = np.prod(u_c.shape[1:])

            t_tensor_reshaped = np.reshape(target_tensor, (-1, shape))
            s_tensor_reshaped = np.reshape(source_tensor, (-1, shape))

            # add penalties on biases
            if backward:
                return np.max(u_c, -1)
            else:
                upper = u_c[:, :, None] - l_c[:, None]
            const = upper.max() - upper.min()
            discard_mask_s = t_tensor_reshaped[:, :, None] * s_tensor_reshaped[:, None, :]

            upper -= (1 - discard_mask_s) * (const + 0.1)

            return np.max(np.max(upper, -2), -1)

        def get_affine_score(
            z_tensor: npt.NDArray[np.float_],
            w_u: npt.NDArray[np.float_],
            b_u: npt.NDArray[np.float_],
            w_l: npt.NDArray[np.float_],
            b_l: npt.NDArray[np.float_],
            source_tensor: npt.NDArray[np.int_],
            target_tensor: Optional[npt.NDArray[np.int_]] = None,
            backward: bool = False,
        ) -> npt.NDArray[np.float_]:
            if target_tensor is None:
                target_tensor = 1 - source_tensor

            n_dim = w_u.shape[1]
            shape = np.prod(b_u.shape[1:])
            w_u_reshaped = np.reshape(w_u, (-1, n_dim, shape, 1))
            w_l_reshaped = np.reshape(w_l, (-1, n_dim, 1, shape))
            b_u_reshaped = np.reshape(b_u, (-1, shape, 1))
            b_l_reshaped = np.reshape(b_l, (-1, 1, shape))

            t_tensor_reshaped = np.reshape(target_tensor, (-1, shape))
            s_tensor_reshaped = np.reshape(source_tensor, (-1, shape))

            if not backward:
                w_u_f = w_u_reshaped - w_l_reshaped
                b_u_f = b_u_reshaped - b_l_reshaped
            else:
                upper = (
                    (np.maximum(w_u, 0.0) * z_tensor[:, 1][:, :, None]).sum(1)
                    + (np.minimum(w_u, 0.0) * z_tensor[:, 0][:, :, None]).sum(1)
                    + b_u
                )

                return upper.max(-1)

            # add penalties on biases
            upper = (
                np.sum(np.maximum(w_u_f, 0) * z_tensor[:, 1, :, None, None], 1)
                + np.sum(np.minimum(w_u_f, 0) * z_tensor[:, 0, :, None, None], 1)
                + b_u_f
            )  # (-1, shape, shape)
            const = upper.max() - upper.min()
            discard_mask_s = t_tensor_reshaped[:, :, None] * s_tensor_reshaped[:, None, :]

            upper -= (1 - discard_mask_s) * (const + 0.1)

            return np.max(np.max(upper, -2), -1)

        if ibp and affine:
            z, u_c, w_u_f, b_u_f, l_c, w_l_f, b_l_f = output[:7]
        elif not ibp and affine:
            z, w_u_f, b_u_f, w_l_f, b_l_f = output[:5]
        elif ibp and not affine:
            u_c, l_c = output[:2]
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if ibp:
            adv_ibp = get_ibp_score(u_c, l_c, source_labels, target_labels, backward=decomon_model.backward_bounds)
        if affine:
            adv_f = get_affine_score(
                z, w_u_f, b_u_f, w_l_f, b_l_f, source_labels, target_labels, backward=decomon_model.backward_bounds
            )

        if ibp and not affine:
            adv_score = adv_ibp
        elif ibp and affine:
            adv_score = np.minimum(adv_ibp, adv_f)
        elif not ibp and affine:
            adv_score = adv_f
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

    if n_split > 1:
        adv_score = np.max(np.reshape(adv_score, (-1, n_split)), -1)

    return adv_score


def check_adv_box(
    model: Union[tf.keras.Model, DecomonModel],
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    source_labels: npt.NDArray[np.int_],
    target_labels: Optional[npt.NDArray[np.int_]] = None,
    batch_size: int = -1,
) -> npt.NDArray[np.float_]:
    """if the output is negative, then it is a formal guarantee that there is no adversarial examples

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        source_labels: the list of label that should be predicted all
            the time in the box (either an integer, either an array that
            can contain multiple source labels for each sample)
        target_labels: the list of label that should never be predicted
            in the box (either an integer, either an array that can
            contain multiple target labels for each sample)
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        numpy array, vector with upper bounds for adversarial attacks
    """
    if np.min(x_max - x_min) < 0:
        raise UserWarning("Inconsistency Error: x_max < x_min")

    # check that the model is a DecomonModel, else do the conversion
    if not isinstance(model, DecomonModel):
        decomon_model = clone(model)
    else:
        assert isinstance(model.perturbation_domain, (BoxDomain, GridDomain))
        decomon_model = model

    ibp = decomon_model.ibp
    affine = decomon_model.affine

    n_split = 1
    n_batch = len(x_min)
    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[2:])
    input_dim = np.prod(input_shape)
    x_reshaped = x_min + 0 * x_min
    x_reshaped = x_reshaped.reshape([-1] + input_shape)

    x_min = x_min.reshape((-1, 1, input_dim))
    x_max = x_max.reshape((-1, 1, input_dim))

    z = np.concatenate([x_min, x_max], 1)
    if isinstance(source_labels, (int, np.int_)):
        source_labels = np.zeros((n_batch, 1)) + source_labels

    if isinstance(source_labels, list):
        source_labels = np.array(source_labels).reshape((n_batch, -1))

    source_labels = source_labels.reshape((n_batch, -1))
    source_labels = source_labels.astype(np.int_)

    if target_labels is not None:
        target_labels = target_labels.reshape((n_batch, -1))
        target_labels = target_labels.astype(np.int_)

    if n_split > 1:
        shape = list(source_labels.shape[1:])
        source_labels = np.reshape(np.concatenate([source_labels[:, None]] * n_split, 1), [-1] + shape)
        if target_labels is not None:
            target_labels = np.reshape(np.concatenate([target_labels[:, None]] * n_split, 1), [-1] + shape)

    if batch_size > 0:
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_min_list = [x_min[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        x_max_list = [x_max[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        source_labels_list = [
            source_labels[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)
        ]
        target_labels_list: List[Optional[npt.NDArray[np.int_]]]
        if (
            (target_labels is not None)
            and (not isinstance(target_labels, int))
            and (str(target_labels.dtype)[:3] != "int")
        ):
            target_labels_list = [
                target_labels[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)
            ]
        else:
            target_labels_list = [target_labels] * (len(x_reshaped) // batch_size + r)

        return np.concatenate(
            [
                check_adv_box(
                    decomon_model, x_min_list[i], x_max_list[i], source_labels_list[i], target_labels_list[i], -1
                )
                for i in range(len(x_min_list))
            ]
        )

    else:
        output = decomon_model.predict(z, verbose=0)

        if not affine:
            # translate  into affine information
            u_c = output[0]
            w_u = 0 * u_c[:, None] + np.zeros((1, input_dim, 1))
            output = [z, w_u, u_c, w_u, output[-1]]
            ibp = False
            affine = True

        def get_affine_sample(
            z_tensor: npt.NDArray[np.float_],
            w_u: npt.NDArray[np.float_],
            b_u: npt.NDArray[np.float_],
            w_l: npt.NDArray[np.float_],
            b_l: npt.NDArray[np.float_],
            source_tensor: npt.NDArray[np.int_],
            target_tensor: Optional[npt.NDArray[np.int_]] = None,
        ) -> npt.NDArray[np.float_]:
            if target_tensor is None:
                target_tensor = 1 - source_tensor

            n_dim = w_u.shape[1]
            shape = np.prod(b_u.shape[1:])
            w_u_reshaped = np.reshape(w_u, (-1, n_dim, shape, 1))
            w_l_reshaped = np.reshape(w_l, (-1, n_dim, 1, shape))
            b_u_reshaped = np.reshape(b_u, (-1, shape, 1))
            b_l_reshaped = np.reshape(b_l, (-1, 1, shape))

            t_tensor_reshaped = np.reshape(target_tensor, (-1, shape))
            s_tensor_reshaped = np.reshape(source_tensor, (-1, shape))

            w_u_f = w_u_reshaped - w_l_reshaped
            b_u_f = b_u_reshaped - b_l_reshaped

            # add penalties on biases
            upper = (
                np.sum(np.maximum(w_u_f, 0) * z_tensor[:, 1, :, None, None], 1)
                + np.sum(np.minimum(w_u_f, 0) * z_tensor[:, 0, :, None, None], 1)
                + b_u_f
            )  # (-1, shape, shape)
            const = upper.max() - upper.min()
            discard_mask_s = t_tensor_reshaped[:, :, None] * s_tensor_reshaped[:, None, :]

            upper -= (1 - discard_mask_s) * (const + 0.1)

            return np.max(np.max(upper, -2), -1)

        if ibp:
            z, u_c, w_u_f, b_u_f, l_c, w_l_f, b_l_f = output[:7]
        else:
            z, w_u_f, b_u_f, w_l_f, b_l_f = output[:5]

        return get_affine_sample(z, w_l_f, b_l_f, w_u_f, b_u_f, source_labels)


#### FORMAL BOUNDS ######
def get_upper_box(
    model: Union[tf.keras.Model, DecomonModel],
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    batch_size: int = -1,
    n_sub_boxes: int = 1,
) -> npt.NDArray[np.float_]:
    """upper bound the maximum of a model in a given box

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        numpy array, vector with upper bounds for adversarial attacks
    """

    return get_range_box(model=model, x_min=x_min, x_max=x_max, batch_size=batch_size, n_sub_boxes=n_sub_boxes)[0]


def get_lower_box(
    model: Union[tf.keras.Model, DecomonModel],
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    batch_size: int = -1,
    n_sub_boxes: int = 1,
) -> npt.NDArray[np.float_]:
    """lower bound the minimum of a model in a given box

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        numpy array, vector with lower bounds for adversarial attacks
    """

    return get_range_box(model=model, x_min=x_min, x_max=x_max, batch_size=batch_size, n_sub_boxes=n_sub_boxes)[1]


def get_range_box(
    model: Union[tf.keras.Model, DecomonModel],
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    batch_size: int = -1,
    n_sub_boxes: int = 1,
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """bounding the outputs of a model in a given box
    if the constant is negative, then it is a formal guarantee that there is no adversarial examples

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        2 numpy array, vector with upper bounds and vector with lower
        bounds
    """
    if np.min(x_max - x_min) < 0:
        raise UserWarning("Inconsistency Error: x_max < x_min")

    # check that the model is a DecomonModel, else do the conversion
    if not (isinstance(model, DecomonModel)):
        decomon_model = clone(model)
    else:
        assert isinstance(model.perturbation_domain, (BoxDomain, GridDomain))
        decomon_model = model

    n_split = 1
    n_batch = len(x_min)

    if n_sub_boxes > 1:
        x_min, x_max = refine_boxes(x_min, x_max, n_sub_boxes)
        # reshape
        n_split = x_min.shape[1]
        shape = list(x_min.shape[2:])
        x_min = np.reshape(x_min, [-1] + shape)
        x_max = np.reshape(x_max, [-1] + shape)

    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[2:])
    input_dim = np.prod(input_shape)
    x_reshaped = x_min + 0 * x_min
    x_reshaped = x_reshaped.reshape([-1] + input_shape)
    x_min = x_min.reshape((-1, 1, input_dim))
    x_max = x_max.reshape((-1, 1, input_dim))

    z = np.concatenate([x_min, x_max], 1)

    if batch_size > 0:
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_min_list = [x_min[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        x_max_list = [x_max[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        results = [get_range_box(decomon_model, x_min_list[i], x_max_list[i], -1) for i in range(len(x_min_list))]

        u_out = np.concatenate([r[0] for r in results])
        l_out = np.concatenate([r[1] for r in results])

    else:
        ibp = decomon_model.ibp
        affine = decomon_model.affine

        output = decomon_model.predict(z, verbose=0)
        shape = list(output[-1].shape[1:])
        output_dim = np.prod(shape)

        if affine:
            if ibp:
                _, u_i, w_u_f, b_u_f, l_i, w_l_f, b_l_f = output[:7]
                if len(u_i.shape) > 2:
                    u_i = np.reshape(u_i, (-1, output_dim))
                    l_i = np.reshape(l_i, (-1, output_dim))
            else:
                _, w_u_f, b_u_f, w_l_f, b_l_f = output[:5]

            # reshape if necessary
            if len(w_u_f.shape) > 3:
                w_u_f = np.reshape(w_u_f, (-1, input_dim, output_dim))
                w_l_f = np.reshape(w_l_f, (-1, input_dim, output_dim))
                b_u_f = np.reshape(b_u_f, (-1, output_dim))
                b_l_f = np.reshape(b_l_f, (-1, output_dim))

            u_f = (
                np.sum(np.maximum(w_u_f, 0) * x_max[:, 0, :, None], 1)
                + np.sum(np.minimum(w_u_f, 0) * x_min[:, 0, :, None], 1)
                + b_u_f
            )
            l_f = (
                np.sum(np.maximum(w_l_f, 0) * x_min[:, 0, :, None], 1)
                + np.sum(np.minimum(w_l_f, 0) * x_max[:, 0, :, None], 1)
                + b_l_f
            )

        else:
            u_i = output[0]
            l_i = output[1]
            if len(u_i.shape) > 2:
                u_i = np.reshape(u_i, (-1, output_dim))
                l_i = np.reshape(l_i, (-1, output_dim))

        if ibp and affine:
            u_out = np.minimum(u_i, u_f)
            l_out = np.maximum(l_i, l_f)
        elif ibp and not affine:
            u_out = u_i
            l_out = l_i
        elif not ibp and affine:
            u_out = u_f
            l_out = l_f
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        #####
        if len(shape) > 1:
            u_out = np.reshape(u_out, [-1] + shape)
            l_out = np.reshape(l_out, [-1] + shape)

    if n_split > 1:
        u_out = np.max(np.reshape(u_out, (-1, n_split)), -1)
        l_out = np.min(np.reshape(l_out, (-1, n_split)), -1)
    return u_out, l_out


# get upper bound of a sample with bounded noise
def get_upper_noise(
    model: Union[tf.keras.Model, DecomonModel],
    x: npt.NDArray[np.float_],
    eps: float,
    p: float = np.inf,
    batch_size: int = -1,
) -> npt.NDArray[np.float_]:
    """upper bound the maximum of a model in an Lp Ball

    Args:
        model: either a Keras model or a Decomon model
        x: numpy array, the example around
        eps: the radius of the ball
        p: the type of Lp norm (p=2, 1, np.inf)
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        numpy array, vector with upper bounds of the range of values
        taken by the model inside the ball
    """

    return get_range_noise(model=model, x=x, eps=eps, p=p, batch_size=batch_size)[0]


# get upper bound of a sample with bounded noise
def get_lower_noise(
    model: Union[tf.keras.Model, DecomonModel],
    x: npt.NDArray[np.float_],
    eps: float,
    p: float = np.inf,
    batch_size: int = -1,
) -> npt.NDArray[np.float_]:
    """lower bound the minimum of a model in an Lp Ball

    Args:
        model: either a Keras model or a Decomon model
        x: numpy array, the example around
        eps: the radius of the ball
        p: the type of Lp norm (p=2, 1, np.inf)
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        numpy array, vector with lower bounds of the range of values
        taken by the model inside the ball
    """

    return get_range_noise(model=model, x=x, eps=eps, p=p, batch_size=batch_size)[1]


# get upper bound of a sample with bounded noise
def get_range_noise(
    model: Union[tf.keras.Model, DecomonModel],
    x: npt.NDArray[np.float_],
    eps: float,
    p: float = np.inf,
    batch_size: int = -1,
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """Bounds the output of a model in an Lp Ball

    Args:
        model: either a Keras model or a Decomon model
        x: numpy array, the example around
        eps: the radius of the ball
        p: the type of Lp norm (p=2, 1, np.inf)
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        2 numpy arrays, vector with upper andlower bounds
    of the range of values taken by the model inside the ball
    """

    # check that the model is a DecomonModel, else do the conversion
    perturbation_domain = BallDomain(p=p, eps=max(0, eps))

    if not isinstance(model, DecomonModel):
        decomon_model = clone(
            model,
            method=ConvertMethod.CROWN_FORWARD_HYBRID,
            perturbation_domain=perturbation_domain,
        )
    else:
        decomon_model = model
        if eps >= 0:
            decomon_model.set_domain(perturbation_domain)

    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[1:])
    input_dim = np.prod(input_shape)
    x_reshaped = x + 0 * x
    x_reshaped = x_reshaped.reshape([-1] + input_shape)

    if batch_size > 0:
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_list = [x_reshaped[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        results = [get_range_noise(decomon_model, x_list[i], eps=eps, p=p, batch_size=-1) for i in range(len(x_list))]

        return np.concatenate([r[0] for r in results]), np.concatenate([r[1] for r in results])

    ibp = decomon_model.ibp
    affine = decomon_model.affine

    output = decomon_model.predict(x_reshaped, verbose=0)
    shape = list(output[-1].shape[1:])
    output_dim = np.prod(shape)

    x_reshaped = x_reshaped.reshape((len(x_reshaped), -1))
    ord = _get_dual_ord(p)

    if affine:
        if ibp:
            _, u_i, w_u_f, b_u_f, l_i, w_l_f, b_l_f = output[:7]
            if len(u_i.shape) > 2:
                u_i = np.reshape(u_i, (-1, output_dim))
                l_i = np.reshape(l_i, (-1, output_dim))
        else:
            _, w_u_f, b_u_f, w_l_f, b_l_f = output[:5]

        # reshape if necessary
        if len(w_u_f.shape) > 3:
            w_u_f = np.reshape(w_u_f, (-1, input_dim, output_dim))
            b_u_f = np.reshape(b_u_f, (-1, output_dim))
            w_l_f = np.reshape(w_l_f, (-1, input_dim, output_dim))
            b_l_f = np.reshape(b_l_f, (-1, output_dim))

        u_f = eps * np.linalg.norm(w_u_f, ord=ord, axis=1) + np.sum(w_u_f * x_reshaped[:, :, None], 1) + b_u_f
        l_f = -eps * np.linalg.norm(w_l_f, ord=ord, axis=1) + np.sum(w_l_f * x_reshaped[:, :, None], 1) + b_l_f

    else:
        u_i = output[0]
        l_i = output[1]
        if len(u_i.shape) > 2:
            u_i = np.reshape(u_i, (-1, output_dim))
            l_i = np.reshape(l_i, (-1, output_dim))
            ######

    if ibp and affine:
        u_out = np.minimum(u_i, u_f)
        l_out = np.maximum(l_i, l_f)
    elif ibp and not affine:
        u_out = u_i
        l_out = l_i
    elif not ibp and affine:
        u_out = u_f
        l_out = l_f
    else:
        raise NotImplementedError("not ibp and not affine not implemented")

    if len(shape) > 1:
        u_out = np.reshape(u_out, [-1] + shape)
        l_out = np.reshape(l_out, [-1] + shape)

    return u_out, l_out


def refine_boxes(
    x_min: npt.NDArray[np.float_], x_max: npt.NDArray[np.float_], n_sub_boxes: int = 10
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    # flatten x_min and x_max
    shape = list(x_min.shape[1:])
    output_dim = np.prod(shape)
    if len(shape) > 1:
        x_min = np.reshape(x_min, (-1, output_dim))
        x_max = np.reshape(x_min, (-1, output_dim))
    # x_min (None, n)
    # x_max (None, n)
    n = x_min.shape[-1]
    X_min = np.zeros((len(x_min), 1, n)) + x_min[:, None]
    X_max = np.zeros((len(x_max), 1, n)) + x_max[:, None]

    def split(
        x_min: npt.NDArray[np.float_], x_max: npt.NDArray[np.float_], j: npt.NDArray[np.int_]
    ) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        n_0 = len(x_min)
        n_k = x_min.shape[1]

        x_min_reshaped = np.zeros((n_0, 2, n_k, n)) + x_min[:, None]
        x_max_reshaped = np.zeros((n_0, 2, n_k, n)) + x_max[:, None]

        index_0 = np.arange(n_0)
        index_k = np.arange(n_k)
        mid_x = (x_min + x_max) / 2.0
        split_value = np.array([mid_x[i, index_k, j[i]] for i in index_0])

        for i in index_0:
            x_min_reshaped[i, 1, index_k, j[i]] = split_value[i]
            x_max_reshaped[i, 0, index_k, j[i]] = split_value[i]

        return x_min_reshaped.reshape((-1, 2 * n_k, n)), x_max_reshaped.reshape((-1, 2 * n_k, n))

    # init
    n_sub = X_min.shape[1]
    while 2 * n_sub <= n_sub_boxes:
        j = np.argmax(X_max - X_min, -1)
        X_min, X_max = split(X_min, X_max, j)
        n_sub = X_min.shape[1]

    if len(shape) > 1:
        X_min = np.reshape(X_min, [-1, n_sub] + shape)
        X_max = np.reshape(X_max, [-1, n_sub] + shape)

    return X_min, X_max


def refine_box(
    func: Callable[..., npt.NDArray[np.float_]],
    model: Union[tf.keras.Model, DecomonModel],
    x_min: npt.NDArray[np.float_],
    x_max: npt.NDArray[np.float_],
    n_split: int,
    source_labels: Optional[npt.NDArray[np.int_]] = None,
    target_labels: Optional[npt.NDArray[np.int_]] = None,
    batch_size: int = -1,
    random: bool = True,
) -> npt.NDArray[np.float_]:
    if func.__name__ not in [
        elem.__name__ for elem in [get_upper_box, get_lower_box, get_adv_box, check_adv_box, get_range_box]
    ]:
        raise NotImplementedError()

    # check that the model is a DecomonModel, else do the conversion
    if not isinstance(model, DecomonModel):
        decomon_model = clone(model)
    else:
        assert isinstance(model.perturbation_domain, (BoxDomain, GridDomain))
        decomon_model = model

    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[2:])
    input_dim = np.prod(input_shape)
    x_min = x_min.reshape((-1, input_dim))
    x_max = x_max.reshape((-1, input_dim))

    n_x = len(x_min)

    # split
    x_min_split = np.zeros((n_x, n_split, input_dim))
    x_max_split = np.zeros((n_x, n_split, input_dim))

    x_min_split[:, 0] = x_min
    x_max_split[:, 0] = x_max

    index_max = np.zeros((n_x, n_split, input_dim))
    # init
    index_max[:, 0] = x_max - x_min

    maximize = True
    if func.__name__ == get_lower_box.__name__:
        maximize = False

    def priv_func(x_min_split: npt.NDArray[np.float_], x_max_split: npt.NDArray[np.float_]) -> npt.NDArray[np.float_]:
        if func.__name__ in [elem.__name__ for elem in [get_upper_box, get_lower_box, get_range_box]]:
            results = func(decomon_model, x_min=x_min_split, x_max=x_max_split, batch_size=batch_size)

            if func.__name__ == get_upper_box.__name__:
                return results.reshape((n_x, n_split, -1))
            if func.__name__ == get_lower_box.__name__:
                return results.reshape((n_x, n_split, -1))

        if func.__name__ in [elem.__name__ for elem in [get_adv_box, check_adv_box]]:
            results = func(
                decomon_model,
                x_min=x_min_split,
                x_max=x_max_split,
                source_labels=source_labels,
                target_labels=target_labels,
                batch_size=batch_size,
            )

            return results.reshape((n_x, n_split))

        raise NotImplementedError()

    for n_i in range(n_x):
        count = 1

        while count < n_split:
            if not random:
                i = int(np.argmax(np.max(index_max[n_i, :count], -1)))
                j = int(np.argmax(index_max[n_i, i]))
            else:
                i = np.random.randint(count - 1)
                j = np.random.randint(input_dim)

            z_min = x_min_split[n_i, i] + 0.0
            z_max = x_max_split[n_i, i] + 0.0
            x_min_split[n_i, count] = z_min + 0.0
            x_max_split[n_i, count] = z_max + 0.0

            x_max_split[n_i, i, j] = (z_min[j] + z_max[j]) / 2.0
            x_min_split[n_i, count, j] = (z_min[j] + z_max[j]) / 2.0

            index_max[n_i, count] = index_max[n_i, i]
            index_max[n_i, i, j] /= 2.0
            index_max[n_i, count, j] /= 2.0

            count += 1

    x_min_split = x_min_split.reshape((-1, input_dim))
    x_max_split = x_max_split.reshape((-1, input_dim))

    results = priv_func(x_min_split, x_max_split)
    if maximize:
        return np.max(results, 1)
    else:
        return np.min(results, 1)


### adversarial robustness Lp norm
def get_adv_noise(
    model: Union[tf.keras.Model, DecomonModel],
    x: npt.NDArray[np.float_],
    source_labels: LabelType,
    eps: float = 0.0,
    p: float = np.inf,
    target_labels: Optional[LabelType] = None,
    batch_size: int = -1,
) -> npt.NDArray[np.float_]:
    """if the output is negative, then it is a formal guarantee that there is no adversarial examples

    Args:
        model: either a Keras model or a Decomon model
        x_min: numpy array for the extremal lower corner of the boxes
        x_max: numpy array for the extremal upper corner of the boxes
        source_labels: the list of label that should be predicted all
            the time in the box (either an integer, either an array that
            can contain multiple source labels for each sample)
        target_labels: the list of label that should never be predicted
            in the box (either an integer, either an array that can
            contain multiple target labels for each sample)
        batch_size: for computational efficiency, one can split the
            calls to minibatches

    Returns:
        numpy array, vector with upper bounds for adversarial attacks
    """

    perturbation_domain = BallDomain(p=p, eps=max(0, eps))

    # check that the model is a DecomonModel, else do the conversion
    if not isinstance(model, DecomonModel):
        decomon_model = clone(model, perturbation_domain=perturbation_domain)
    else:
        decomon_model = model
        if eps >= 0:
            decomon_model.set_domain(perturbation_domain)

    eps = model.perturbation_domain.eps

    # reshape x_mmin, x_max
    input_shape = list(decomon_model.input_shape[1:])
    x_reshaped = x + 0 * x
    x_reshaped = x_reshaped.reshape([-1] + input_shape)
    n_split = 1
    n_batch = len(x_reshaped)

    input_shape = list(decomon_model.input_shape[1:])
    x_reshaped = x + 0 * x
    x_reshaped = x_reshaped.reshape([-1] + input_shape)

    if isinstance(source_labels, (int, np.int_)):
        source_labels = np.zeros((n_batch, 1), dtype=np.int_) + source_labels
    source_labels = np.array(source_labels).reshape((n_batch, -1)).astype(np.int_)

    if target_labels is not None:
        if isinstance(target_labels, (int, np.int_)):
            target_labels = np.zeros((n_batch, 1), dtype=np.int_) + source_labels
        target_labels = np.array(target_labels).reshape((n_batch, -1)).astype(np.int_)

    if batch_size > 0:
        # split
        r = 0
        if len(x_reshaped) % batch_size > 0:
            r += 1
        x_list = [x_reshaped[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)]
        source_labels_list = [
            source_labels[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)
        ]
        target_labels_list: List[Optional[npt.NDArray[np.int_]]]
        if (
            (target_labels is not None)
            and (not isinstance(target_labels, int))
            and (str(target_labels.dtype)[:3] != "int")
        ):
            target_labels_list = [
                target_labels[batch_size * i : batch_size * (i + 1)] for i in range(len(x_reshaped) // batch_size + r)
            ]
        else:
            target_labels_list = [target_labels] * (len(x_reshaped) // batch_size + r)

        results = [
            get_adv_noise(
                decomon_model,
                x_list[i],
                source_labels=source_labels_list[i],
                eps=eps,
                p=p,
                target_labels=target_labels_list[i],
                batch_size=-1,
            )
            for i in range(len(x_list))
        ]

        return np.concatenate(results)
    else:
        ibp = decomon_model.ibp
        affine = decomon_model.affine
        output = decomon_model.predict(x_reshaped, verbose=0)

        def get_ibp_score(
            u_c: npt.NDArray[np.float_],
            l_c: npt.NDArray[np.float_],
            source_tensor: npt.NDArray[np.int_],
            target_tensor: Optional[npt.NDArray[np.int_]] = None,
        ) -> npt.NDArray[np.float_]:
            if target_tensor is None:
                target_tensor = 1 - source_tensor

            shape = np.prod(u_c.shape[1:])

            t_tensor_reshaped = np.reshape(target_tensor, (-1, shape))
            s_tensor_reshaped = np.reshape(source_tensor, (-1, shape))

            # add penalties on biases
            upper = u_c[:, :, None] - l_c[:, None]
            const = upper.max() - upper.min()
            discard_mask_s = t_tensor_reshaped[:, :, None] * s_tensor_reshaped[:, None, :]

            upper -= (1 - discard_mask_s) * (const + 0.1)

            return np.max(np.max(upper, -2), -1)

        def get_affine_score(
            z_tensor: npt.NDArray[np.float_],
            w_u: npt.NDArray[np.float_],
            b_u: npt.NDArray[np.float_],
            w_l: npt.NDArray[np.float_],
            b_l: npt.NDArray[np.float_],
            source_tensor: npt.NDArray[np.int_],
            target_tensor: Optional[npt.NDArray[np.int_]] = None,
        ) -> npt.NDArray[np.float_]:
            if target_tensor is None:
                target_tensor = 1 - source_tensor

            n_dim = w_u.shape[1]
            shape = np.prod(b_u.shape[1:])
            w_u_reshaped = np.reshape(w_u, (-1, n_dim, shape, 1))
            w_l_reshaped = np.reshape(w_l, (-1, n_dim, 1, shape))
            b_u_reshaped = np.reshape(b_u, (-1, shape, 1))
            b_l_reshaped = np.reshape(b_l, (-1, 1, shape))

            t_tensor_reshaped = np.reshape(target_tensor, (-1, shape))
            s_tensor_reshaped = np.reshape(source_tensor, (-1, shape))

            w_u_f = w_u_reshaped - w_l_reshaped
            b_u_f = b_u_reshaped - b_l_reshaped

            # add penalties on biases

            # compute upper with lp norm

            if len(z_tensor.shape) > 2:
                z_tensor = np.reshape(z_tensor, (len(z_tensor), -1))

            upper_0 = np.sum(w_u_f * z_tensor[:, :, None, None], 1) + b_u_f
            # compute dual norm
            if p == 2:
                upper_1 = eps * np.sum(w_u_f**2, 1)
            elif p == 1:
                upper_1 = eps * np.max(np.abs(w_u_f), 1)
            elif p == np.inf:
                upper_1 = eps * np.sum(np.abs(w_u_f), 1)
            else:
                raise ValueError(f"p must be equal to 1, 2, or np.inf, unknown value {p}.")

            upper = upper_0 + upper_1

            const = upper.max() - upper.min()
            discard_mask_s = t_tensor_reshaped[:, :, None] * s_tensor_reshaped[:, None, :]
            upper -= (1 - discard_mask_s) * (const + 0.1)

            return np.max(np.max(upper, -2), -1)

        if ibp and affine:
            z, u_c, w_u_f, b_u_f, l_c, w_l_f, b_l_f = output[:7]
        if not ibp and affine:
            z, w_u_f, b_u_f, w_l_f, b_l_f = output[:5]
        if ibp and not affine:
            u_c, l_c = output[:2]
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if ibp:
            adv_ibp = get_ibp_score(u_c, l_c, source_labels, target_labels)
        if affine:
            adv_f = get_affine_score(z, w_u_f, b_u_f, w_l_f, b_l_f, source_labels, target_labels)

        if ibp and not affine:
            adv_score = adv_ibp
        elif ibp and affine:
            adv_score = np.minimum(adv_ibp, adv_f)
        elif not ibp and affine:
            adv_score = adv_f
        else:
            raise NotImplementedError("not ibp and not affine not implemented")

        if n_split > 1:
            adv_score = np.max(np.reshape(adv_score, (-1, n_split)), -1)

        return adv_score
